#!/usr/bin/env python

# std
from __future__ import annotations
from dataclasses import dataclass
import logging
import sys

# pkg
from cosmofy.args import global_options
from cosmofy.args import GlobalArgs
from cosmofy.baton import arg
from cosmofy.baton import Command
from cosmofy.fs.add import cmd as add
from cosmofy.fs.args import cmd as args
from cosmofy.fs.cat import cmd as cat
from cosmofy.fs.ls import cmd as ls
from cosmofy.fs.rm import cmd as rm


log = logging.getLogger(__name__)

usage = f"""\
Cosmopolitan file system tool.

Usage: cosmofy fs [OPTIONS] <COMMAND>

Commands:
  ls                        list files in bundle
  cat                       print file contents
  add                       add files to bundle
  rm                        remove files from bundle
  args                      get/set special `.args` file in bundle

{global_options}
"""


@dataclass
class Args(GlobalArgs):
    command: str = arg("", positional=True)  # optional so we can show usage


def run(_: Args) -> int:
    """Main entry point for `cosmofy fs`."""
    # NOTE: only called when there was no subcommand found
    cmd.show_usage()
    return 0


cmd = Command(
    "cosmofy.fs",
    Args,
    run,
    usage,
    {
        "ls": ls,
        "cat": cat,
        "add": add,
        "rm": rm,
        "args": args,
    },
)

if __name__ == "__main__":
    sys.exit(cmd.main())



#!/usr/bin/env python

# std
from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
from zipfile import is_zipfile
import logging
import sys

# pkg
from cosmofy.args import global_options
from cosmofy.args import GlobalArgs
from cosmofy.baton import Command
from cosmofy.updater.run import self_update


log = logging.getLogger(__name__)

usage = f"""\
Update `cosmofy`.

Usage: cosmofy self update [OPTIONS]

{global_options}
"""


@dataclass
class Args(GlobalArgs):
    __doc__ = usage


def run(args: Args) -> int:
    """Entry point for `cosmofy self update`."""
    args.setup_logger()
    try:
        if not is_zipfile(sys.executable):
            err = "`cosmofy` was not installed in a Cosmopolitan Python bundle"
            err += "\n  tip: see https://github.com/metaist/cosmofy#install"
            raise ValueError(err)

        self_update(Path(sys.executable))
    except Exception as e:
        args.show_error(log, e)
        return 2
    return 0


cmd = Command("cosmofy.self.update", Args, run)

if __name__ == "__main__":
    sys.exit(cmd.main())



#!/usr/bin/env python

# std
from __future__ import annotations
from dataclasses import dataclass
import logging
import sys

# pkg
from cosmofy.args import global_options
from cosmofy.args import GlobalArgs
from cosmofy.baton import arg
from cosmofy.baton import Command

from .add import cmd as add
from .check import cmd as check
from .remove import cmd as remove


log = logging.getLogger(__name__)

usage = f"""\
EXPERIMENTAL: Manage a bundle's self-updater.

Usage: cosmofy updater [OPTIONS] <COMMAND>

Commands:
  add                       add self-updater to a bundle
  remove                    remove self-updater from a bundle
  check                     check if the bundle has updates

{global_options}
"""


@dataclass
class Args(GlobalArgs):
    __doc__ = usage
    command: str = arg("", positional=True)  # optional so we can show usage


def run(_: Args) -> int:
    """Main entry point for `cosmofy updater`."""
    # NOTE: only called when there was no subcommand found
    cmd.show_usage()
    return 0


cmd = Command(
    "cosmofy.updater",
    Args,
    run,
    subcommands={
        "add": add,
        "remove": remove,
        "rm": remove,
        "check": check,
    },
)

if __name__ == "__main__":
    sys.exit(cmd.main())



#!/usr/bin/env python

# std
from __future__ import annotations
from dataclasses import dataclass
from importlib.metadata import distributions
from os import environ as ENV
from pathlib import Path
from typing import Literal
import json
import logging
import os
import shlex
import shutil
import stat
import subprocess
import sys
import tempfile
import zipfile

# pkg
from cosmofy.args import global_options
from cosmofy.args import GlobalArgs
from cosmofy.baton import arg
from cosmofy.baton import Command
from cosmofy.fs.add import add_path
from cosmofy.fs.args import set_args
from cosmofy.updater.downloader import download
from cosmofy.updater.downloader import download_if_newer
from cosmofy.updater.pythonoid import python_call
from cosmofy.updater.receipt import get_version
from cosmofy.zipfile2 import ZipFile2


log = logging.getLogger(__name__)


DEFAULT_PYTHON_URL = "https://cosmo.zip/pub/cosmos/bin/python"
"""Default URL to download python from."""

COSMOFY_PYTHON_URL = ENV.get("COSMOFY_PYTHON_URL", "")
"""URL to download python from."""

COSMOFY_NO_CACHE = ENV.get("COSMOFY_NO_CACHE", "")
"""Whether to disable cache."""

DEFAULT_CACHE_DIR = Path.home() / ".cache" / "cosmofy"
"""Default cache directory."""

COSMOFY_CACHE_DIR = ENV.get("COSMOFY_CACHE_DIR", "")
"""Path to cache directory."""

BUNDLE_EXCLUDE = {
    "direct_url.json",
    "uv_build.json",
    "uv_cache.json",
}
"""Exclude uv artifacts."""


usage = f"""\
Build a Python project into a Cosmopolitan bundle.

Usage: cosmofy bundle [OPTIONS]

Input options:
      --entry <NAME>        `console_script` entry points to bundle
      --script <PATH>       paths to bundle

      If neither --entry nor --script is specified, all entry points
      will be bundled. If both are specified, entries will be bundled first.

      --python-url <URL>    URL from which to download Cosmopolitan Python
                            [default: {DEFAULT_PYTHON_URL}]
                            [env: COSMOFY_PYTHON_URL={COSMOFY_PYTHON_URL}]

Output options:
  -o, --output-dir <PATH>   output directory
                            [default: project-root/dist]

Cache options:
  -n, --no-cache            do not read or save to the cache
                            [env: COSMOFY_NO_CACHE={COSMOFY_NO_CACHE}]
      --cache-dir <PATH>    store Cosmopolitan Python downloads
                            [default: {str(DEFAULT_CACHE_DIR).replace(str(Path.home()), "~")}]
                            [env: COSMOFY_CACHE_DIR={COSMOFY_CACHE_DIR}]

{global_options}
"""


@dataclass
class Args(GlobalArgs):
    __doc__ = usage

    # input
    entry: list[str] = arg(list)
    script: list[Path] = arg(list)
    python_url: str = arg(DEFAULT_PYTHON_URL, env="COSMOFY_PYTHON_URL")

    # output
    output_dir: Path | None = arg(None)

    # cache
    no_cache: bool = arg(False, short="-n", env="COSMOFY_NO_CACHE")
    cache_dir: Path = arg(DEFAULT_CACHE_DIR, env="COSMOFY_CACHE_DIR")


def open_zip(file: str | Path, *, mode: Literal["r", "w", "x", "a"] = "r") -> ZipFile2:
    """Wrapper for creating a zip file."""
    return ZipFile2(file, mode, compression=zipfile.ZIP_DEFLATED, compresslevel=9)


def ensure_uv(cmd: str = "uv") -> bool:
    """Raise an error if `uv` is not installed."""
    result = shutil.which(cmd)
    if not result:
        err = f"cannot find command: `{cmd}`"
        err += "\n  tip: see https://docs.astral.sh/uv/getting-started/installation/"
        raise FileNotFoundError(err)
    log.debug(f"`uv` installed at {result}")
    return True


def find_project_root(start: Path | None = None) -> Path:
    """Return the first path that has a `pyproject.toml` in it."""
    start = (start or Path.cwd()).resolve()
    for d in (start, *start.parents):
        if (d / "pyproject.toml").is_file():
            return d
    raise FileNotFoundError("no `pyproject.toml` found")


def venv_site_packages(venv: Path) -> Path:
    """Return the path to `site-packages`."""
    # posix: venv/lib/pythonX.Y/site-packages
    lib = venv / "lib"
    if lib.is_dir():
        candidates = sorted(lib.glob("python*/site-packages"))
        if candidates:
            return candidates[-1]
    # windows: venv/Lib/site-packages
    sp = venv / "Lib" / "site-packages"
    if sp.is_dir():
        return sp
    raise FileNotFoundError(f"cannot find `site-packages` in {venv}")


def console_scripts_from_venv(pkg: str, venv: Path) -> dict[str, str]:
    """Get `console_scripts` for a specific package from a `venv`."""
    sp = venv_site_packages(venv)
    pkg_normalized = pkg.lower().replace("_", "-").replace(".", "-")  # PEP 503
    for dist in distributions(path=[str(sp)]):
        dist_name = dist.metadata["Name"].lower().replace("_", "-").replace(".", "-")
        if dist_name == pkg_normalized:
            return {
                ep.name: ep.value
                for ep in dist.entry_points
                if ep.group == "console_scripts"
            }
    return {}


class Bundler:
    args: Args
    banner: str

    bundle: ZipFile2

    def __init__(self, args: Args):
        """Construct a bundler."""
        self.args = args
        self.banner = args.banner

    # File System

    def fs_copy(self, src: Path, dest: Path) -> Path:
        """Copy a file from `src` to `dest`."""
        log.debug(f"{self.banner}copy: {src} to {dest}")
        if self.args.for_real:
            shutil.copy(src, dest)
        return dest

    def fs_set_executable(self, src: Path) -> Path:
        """Set the executable bit on a file."""
        log.debug(f"{self.banner}chmod +x {src}")
        if self.args.for_real:
            mode = src.stat().st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
            src.chmod(mode)
        return src

    # cosmopolitan python

    def from_cache(self, src: Path, dest: Path) -> Path:
        """Copy the archive from cache."""
        log.debug(f"{self.banner}download (if newer): {self.args.python_url}")
        if self.args.for_real:
            download_if_newer(self.args.python_url, src)
        return self.fs_set_executable(self.fs_copy(src, dest))

    def from_download(self, dest: Path) -> Path:
        """Download archive."""
        log.debug(f"{self.banner}download (fresh): {self.args.python_url} to {dest}")
        if self.args.for_real:
            download(self.args.python_url, dest)
        return self.fs_set_executable(dest)

    def get_cosmo_python(self, dest: Path) -> Path:
        """Return a `Path` a Cosmopolitan Python executable."""
        if self.args.no_cache:  # fresh
            return self.from_download(dest)
        return self.from_cache(self.args.cache_dir / "python", dest)

    # uv

    def uv_version(self) -> tuple[str, str]:
        """Return package name and version."""
        cmd = ["uv", "version", "--no-build", "--output-format", "json"]
        log.debug(f"{self.banner}run: {shlex.join(cmd)}")
        if self.args.for_real:
            data = json.loads(subprocess.check_output(cmd, text=True))
            name = data.get("package_name", "")
            ver = data.get("version", "")
            return name, ver
        return "", ""

    def uv_sync(
        self,
        *,
        pkg: str,
        version: str,
        venv: Path | None,
        script: Path | None = None,
    ) -> Path:
        """Return the venv used during `uv sync`."""
        args: list[str] = [
            "uv",
            "sync",
            "--python",
            shlex.quote(version),
            "--no-editable",
            "--exact",
            "--output-format",
            "json",
        ]
        env: dict[str, str] = {**ENV}
        if venv:
            env["VIRTUAL_ENV"] = str(venv)
            env["UV_PROJECT_ENVIRONMENT"] = str(venv)

        if script:
            args.extend(["--script", str(script)])
        else:
            args.extend(["--no-default-groups", "--reinstall-package", pkg])

        if self.args.verbosity > 0:
            args.append(f"-{'v' * self.args.verbosity}")
        elif self.args.verbosity < 0:
            args.append(f"-{'q' * abs(self.args.verbosity)}")
        if self.args.dry_run:
            args.append("--dry-run")

        log.debug(f"{self.banner}run: {shlex.join(args)}")
        if self.args.for_real:
            out = subprocess.check_output(args, text=True, env=env)
            data = json.loads(out)
            log.debug(data)

            # NOTE: `uv sync --script` doesn't respect environment variables.
            venv = Path(data.get("sync", {}).get("environment", {}).get("path", ""))
        else:  # dummy value
            venv = Path()

        log.info(f"{self.banner}uv sync")
        return venv

    # bundle

    def bundle_venv(self, bundle: ZipFile2, venv: Path) -> ZipFile2:
        """Bundle a `venv` into `bundle`."""
        pkgs = venv_site_packages(venv).parent
        for dirname, _, files in os.walk(pkgs):
            for f in files:
                if f in BUNDLE_EXCLUDE and ".dist-info" in str(dirname):
                    continue
                src = Path(dirname) / f
                rel = src.relative_to(pkgs)
                dest = "/".join(("Lib",) + rel.parts)
                add_path(bundle, src, dest, dry_run=self.args.dry_run)
        return bundle

    def bundle_entry_point(
        self,
        src: Path,
        dest: Path,
        *,
        entry_point: str,
        venv: Path | None = None,
    ) -> Path:
        """Return `Path` to the build entry point."""
        self.fs_copy(src, dest)
        with open_zip(dest, mode="a") as bundle:
            if venv:
                self.bundle_venv(bundle, venv)
            set_args(bundle, python_call(entry_point), dry_run=self.args.dry_run)

        log.info(f"bundled: {dest}")
        return dest

    def bundle_entry_points(
        self,
        pkg: str,
        version: str,
        venv: Path,
        cosmo_python: Path,
        output_dir: Path,
    ) -> dict[str, Path]:
        """Return mapping of entry point names to their bundle paths."""
        result: dict[str, Path] = {}

        venv = self.uv_sync(pkg=pkg, version=version, venv=venv)
        # have all deps built

        entry_points = console_scripts_from_venv(pkg, venv)
        if not self.args.entry:
            self.args.entry = list(entry_points.keys())

        for name in self.args.entry:
            if name not in entry_points:
                err = f"could not find entry point: '{name}'"
                err += "\n  tip: look at `[project.scripts]` in `pyproject.toml`"
                raise ValueError(err)
        # all entry points exist

        name = self.args.entry[0]
        first = self.bundle_entry_point(
            src=cosmo_python,
            dest=output_dir / name,
            venv=venv,
            entry_point=entry_points[name],
        )
        result[name] = first
        for name in self.args.entry[1:]:  # bundle rest by replacing .args
            result[name] = self.bundle_entry_point(
                first,
                output_dir / name,
                entry_point=entry_points[name],
            )
        # all entry points built
        return result

    def bundle_script(
        self,
        version: str,
        venv: Path | None,
        cosmo_python: Path,
        output_dir: Path | None,
        script: Path,
    ) -> Path:
        """Bundle an individual script."""
        if not script.is_file():
            raise FileNotFoundError(f"cannot find script file: {script}")

        dest = self.fs_copy(cosmo_python, (output_dir or script.parent) / script.stem)
        with open_zip(dest, mode="a") as bundle:
            self.bundle_venv(
                bundle,
                self.uv_sync(
                    pkg=script.name, version=version, venv=venv, script=script
                ),
            )
            add_path(bundle, script, script.name, dry_run=self.args.dry_run)
            set_args(bundle, script.name, dry_run=self.args.dry_run)

        log.info(f"bundled: {dest}")
        return script

    def run(self) -> None:
        """Build a venv and bundle it into a Cosmopolitan Python executable."""
        with tempfile.TemporaryDirectory(prefix="cosmofy-python-") as cosmo_temp:
            log.debug(f"temp dir={cosmo_temp}")
            cosmo_python = self.get_cosmo_python(Path(cosmo_temp) / "python")
            version = get_version(cosmo_python)
            if not version:
                raise ValueError("could not get Cosmopolitan Python version")
            # have cosmo python + version

            with tempfile.TemporaryDirectory(prefix="cosmofy-venv-") as venv_temp:
                log.debug(f"temp dir={venv_temp}")
                venv = Path(venv_temp)
                if self.args.entry or (not self.args.entry and not self.args.script):
                    if not self.args.output_dir:
                        self.args.output_dir = find_project_root() / "dist"
                    self.args.output_dir.mkdir(parents=True, exist_ok=True)
                    # have an output dir

                    pkg, pkg_ver = self.uv_version()
                    self.bundle_entry_points(
                        pkg=pkg,
                        version=version,
                        venv=venv,
                        cosmo_python=cosmo_python,
                        output_dir=self.args.output_dir,
                    )
                # all entry points built

                for script in self.args.script:
                    self.bundle_script(
                        version,
                        venv,
                        cosmo_python,
                        self.args.output_dir,
                        script,
                    )
                # all scripts built


def run(args: Args) -> int:
    """Main entry point for `cosmofy bundle`."""
    args.setup_logger()
    try:
        ensure_uv()
        Bundler(args).run()
    except Exception as e:
        args.show_error(log, e)
        return 2
    return 0


cmd = Command("cosmofy.bundle", Args, run)

if __name__ == "__main__":
    sys.exit(cmd.main())



"""cosmofy: Cosmopolitan Python Bundler"""

# std
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version

try:
    __version__ = version("cosmofy")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "0.0.0"

__pubdate__ = "2025-12-24T18:35:27Z"



#!/usr/bin/env python

# std
from __future__ import annotations
from dataclasses import dataclass
from getpass import getpass
import logging
import sys

# pkg
from cosmofy.args import common_args
from cosmofy.args import CommonArgs
from cosmofy.args import global_options
from cosmofy.baton import arg
from cosmofy.baton import Command
from cosmofy.zipfile2 import ZipFile2

from . import expand_glob


log = logging.getLogger(__name__)

usage = f"""\
Print contents of a file within a Cosmopolitan bundle.

Usage: cosmofy fs cat <BUNDLE> <FILE>... [OPTIONS]

Arguments:
{common_args}
  <FILE>...                 one or more file patterns to show

  tip: Use `--` to separate options from filenames that start with `-`
  Example: cosmofy fs cat bundle.zip -- -weird-filename.txt

Options:
  -p, --prompt              prompt for a decryption password

{global_options}
"""


@dataclass
class Args(CommonArgs):
    __doc__ = usage
    file: list[str] = arg(list, positional=True, required=True)
    """Patterns of files to print."""

    prompt: bool = arg(False, short="-p")
    """Whether to prompt for a password."""


def show_files(bundle: ZipFile2, args: Args) -> None:
    """Print out the contents of the files."""
    banner = args.banner

    password: bytes | None = None
    if args.for_real and args.prompt:
        password = getpass().encode("utf-8")

    names = bundle.namelist()
    for pat in args.file:
        for name in expand_glob(names, pat):
            if name.endswith("/"):  # ignore directories
                continue
            if args.for_real:  # write bytes straight to the terminal
                sys.stdout.buffer.write(bundle.read(name, password))
                sys.stdout.buffer.flush()
            else:
                log.info(f"{banner}show: {name}")


def run(args: Args) -> int:
    """Entry point for `cosmofy fs cat`."""
    args.setup_logger()
    try:
        assert args.ensure_bundle() and args.bundle
        with ZipFile2(args.bundle) as bundle:
            show_files(bundle, args)
    except Exception as e:
        args.show_error(log, e)
        return 2
    return 0


cmd = Command("cosmofy.fs.cat", Args, run)

if __name__ == "__main__":
    sys.exit(cmd.main())



#!/usr/bin/env python

# std
from __future__ import annotations
from dataclasses import dataclass
import logging
import sys

# pkg
from cosmofy.args import global_options
from cosmofy.args import GlobalArgs
from cosmofy.baton import arg
from cosmofy.baton import Command
from cosmofy.self.update import cmd as update
from cosmofy.self.version import cmd as version

log = logging.getLogger(__name__)

usage = f"""\
Manage the `cosmofy` executable.

Usage: cosmofy self [OPTIONS] <COMMAND>

Commands:
  update                    update `cosmofy`
  version                   display `cosmofy`'s version

{global_options}
"""


@dataclass
class Args(GlobalArgs):
    __doc__ = usage
    command: str = arg("", positional=True)  # optional so we can show usage


def run(_: Args) -> int:
    """Main entry point for `cosmofy self`."""
    # NOTE: only called when there was no subcommand found
    cmd.show_usage()
    return 0


cmd = Command(
    "cosmofy.self",
    Args,
    run,
    subcommands={
        "update": update,
        "version": version,
    },
)

if __name__ == "__main__":
    sys.exit(cmd.main())



#!/usr/bin/env python

# std
from __future__ import annotations
from dataclasses import dataclass
from importlib import metadata
from importlib.metadata import PathDistribution
from os import environ as ENV
from pathlib import Path
from shutil import which
from zipfile import is_zipfile
from zipfile import Path as ZipPath
import logging
import re
import shlex
import subprocess
import sys

# pkg
from cosmofy.args import common_args
from cosmofy.args import CommonArgs
from cosmofy.args import get_banner
from cosmofy.args import global_options
from cosmofy.baton import arg
from cosmofy.baton import Command
from cosmofy.fs.add import add_data
from cosmofy.fs.add import add_path
from cosmofy.fs.args import get_args
from cosmofy.fs.args import set_args
from cosmofy.zipfile2 import ZipFile2

from .pythonoid import PythonArgs
from .receipt import get_version
from .receipt import Receipt


log = logging.getLogger(__name__)

RECEIPT_URL = ENV.get("RECEIPT_URL", "")
"""Override the embedded URL for downloading update metadata."""

RELEASE_URL = ENV.get("RELEASE_URL", "")
"""Override the published URL for downloading the update."""

COSMOFY_NO_COPY = ENV.get("COSMOFY_NO_COPY", "")
"""Skip copying `cosmofy` to bundle."""

COSMOFY_NO_ARGS = ENV.get("COSMOFY_NO_ARGS", "")
"""Skip setting `.args` in bundle."""

ARGS_PREFIX = "-m cosmofy.updater.run"
"""Prefix for updater `.args`"""

PATH_COSMOFY = "Lib/site-packages/cosmofy"
"""Path within zip file to cosmofy package."""

PATH_RECEIPT = f"{PATH_COSMOFY}/.cosmofy.json"
"""Path within the zip file to the local receipt."""

usage = f"""\
EXPERIMENTAL: Add self-updater to a cosmofy bundle.

Usage: cosmofy updater add <BUNDLE> [OPTIONS]

Arguments:
{common_args}

Receipt options:
      --receipt <PATH>      output path to the JSON receipt
                            default is <BUNDLE> + `.json`
      --receipt-url <URL>   URL to the published receipt
                            default is --release-url + `.json`
                            [env: RECEIPT_URL={RECEIPT_URL}]
      --release-url <URL>   URL to the file to download
                            default is --receipt-url without `.json`
                            [env: RELEASE_URL={RELEASE_URL}]
      --release-version <STRING>
                            release version
                            default is $(<BUNDLE> --version)

Process options:
      --no-copy             skip copying `cosmofy` code
                            (e.g., its already a dependency)
                            [env: COSMOFY_NO_COPY={COSMOFY_NO_COPY}]
      --no-args             skip setting `.args`
                            [env: COSMOFY_NO_ARGS={COSMOFY_NO_ARGS}]

read more: https://github.com/metaist/cosmofy#self-updater

{global_options}
"""


@dataclass
class Args(CommonArgs):
    __doc__ = usage
    receipt: Path | None = arg(None)
    receipt_url: str = arg(RECEIPT_URL)
    release_url: str = arg(RELEASE_URL)
    release_version: str = arg("")
    no_copy: bool = arg(False, env="COSMOFY_NO_COPY")
    no_args: bool = arg(False, env="COSMOFY_NO_ARGS")


def get_github_download(name: str) -> str:
    """Return GitHub download url based on the current repo."""
    if not which("git"):
        return ""  # git not installed

    try:
        url = subprocess.check_output(
            ["git", "remote", "get-url", "origin"], text=True, stderr=subprocess.STDOUT
        ).strip()
    except subprocess.CalledProcessError:
        return ""  # not in a git repo or other error

    # SSH forms:
    #   git@github.com:owner/repo.git
    #   ssh://git@github.com/owner/repo.git
    m = re.match(
        r"^(?:ssh://)?git@github\.com[:/](?P<owner>[^/]+)/(?P<repo>[^/]+?)(?:\.git)?/?$",
        url,
    )
    if not m:
        # HTTPS forms:
        #   https://github.com/owner/repo.git
        #   https://github.com/owner/repo
        m = re.match(
            r"^https://github\.com/(?P<owner>[^/]+)/(?P<repo>[^/]+?)(?:\.git)?/?$", url
        )

    if not m:
        return ""

    owner = m.group("owner")
    repo = m.group("repo")
    return f"https://github.com/{owner}/{repo}/releases/latest/download/{name}"


def copy_data(
    src: ZipPath, dest: ZipFile2, *, dry_run: bool = False, level: int = logging.INFO
) -> None:
    """Copy data from one zip file to another."""
    if src.is_dir():
        for item in src.iterdir():
            copy_data(item, dest, dry_run=dry_run, level=level)
    elif src.is_file():
        add_data(
            dest,
            src.read_bytes(),
            src.at,  # type: ignore
            force=True,
            dry_run=dry_run,
            level=level,
        )


def from_cosmo(
    bundle: ZipFile2,
    src_path: Path,
    meta_path: Path,
    *,
    dry_run: bool = False,
) -> None:
    """Copy `cosmofy` from the current executable."""
    root = ZipPath(sys.executable)
    copy_data(
        root / "/".join(src_path.parts[2:]),
        bundle,
        dry_run=dry_run,
        level=logging.DEBUG,
    )
    copy_data(
        root / "/".join(meta_path.parts[2:]),
        bundle,
        dry_run=dry_run,
        level=logging.DEBUG,
    )


def from_venv(
    bundle: ZipFile2,
    src_path: Path,
    meta_path: Path,
    *,
    dry_run: bool = False,
) -> None:
    """Copy `cosmofy` from the filesystem into `bundle`."""
    add_path(
        bundle,
        src_path,
        PATH_COSMOFY,
        force=True,
        dry_run=dry_run,
        level=logging.DEBUG,
    )

    meta_dest = f"Lib/site-packages/{meta_path.name}"
    add_path(
        bundle,
        meta_path,
        meta_dest,
        force=True,
        dry_run=dry_run,
        level=logging.DEBUG,
    )


def copy_cosmofy(bundle: ZipFile2, *, dry_run: bool = False) -> None:
    """Copy `cosmofy` package into `bundle`."""
    dist = metadata.distribution("cosmofy")
    src_path = Path(__file__).parent.parent
    site_packages = src_path.parent

    dist_info_name = f"{dist.name}-{dist.version}.dist-info"
    meta_path = site_packages / dist_info_name

    if not meta_path.is_dir():
        if hasattr(dist, "_path") and isinstance(dist, PathDistribution):
            meta_path = Path(str(dist._path))
        else:
            raise FileNotFoundError(f"could not location {dist_info_name}")

    if is_zipfile(sys.executable):
        log.debug("inside cosmo python")
        from_cosmo(bundle, src_path, meta_path, dry_run=dry_run)
    else:
        log.debug("inside python venv")
        from_venv(bundle, src_path, meta_path, dry_run=dry_run)


def add_arg_prefix(python_args: str) -> str:
    """Return updated args."""
    args = python_args.strip()
    try:
        PythonArgs.parse(shlex.split(args))
        if args.startswith(ARGS_PREFIX):  # idempotent
            return args
        return f"{ARGS_PREFIX} {args}".strip()
    except ValueError as e:
        err = f"unsupported python args: '{args}'"
        err += "\n  tip: see https://github.com/metaist/cosmofy#supported-python-cli"
        raise ValueError(err) from e


def write_receipt(
    path: Path,
    bundle: ZipFile2,
    *,
    output: Path,
    receipt_url: str,
    release_url: str,
    release_version: str,
    dry_run: bool = False,
) -> Receipt:
    """Write remote receipt."""
    for_real = not dry_run
    banner = get_banner(dry_run)

    receipt = Receipt(
        receipt_url=receipt_url,
        release_url=release_url,
        version=release_version,
    )
    data = str(receipt)
    log.debug(data)
    if not receipt.is_valid():
        raise ValueError("embedded receipt must be valid")

    add_data(bundle, data, PATH_RECEIPT, force=True, dry_run=dry_run)

    receipt.update_from(
        Receipt.from_path(path, version=release_version),
        "algo",
        "hash",
        "version",
        kind="published",
    )
    data = str(receipt)
    log.debug(data)
    if not receipt.is_valid():
        raise ValueError("published receipt must be valid")

    if for_real:
        output.write_text(data)
    log.info(f"{banner}wrote: {output}")
    return receipt


def run(args: Args) -> int:
    """Entry point for `cosmofy updater add`."""
    args.setup_logger()
    log.warning("EXPERIMENTAL: `cosmofy updater` operations are experimental")
    try:
        assert args.ensure_bundle() and args.bundle
        if not args.receipt:
            args.receipt = args.bundle.with_suffix(f"{args.bundle.suffix}.json")
            log.info(f"infer --receipt: {args.receipt}")

        if not args.release_url:  # try github download
            args.release_url = get_github_download(args.bundle.name)
            log.info(f"infer --release-url: {args.release_url}")
        if not args.receipt_url and not args.release_url:
            err = "at least one of --receipt-url or --release-url is required"
            err += "\n  note: auto-detection only works for GitHub repositories"
            raise ValueError(err)
        if not args.receipt_url:
            args.receipt_url = args.release_url + ".json"
            log.info(f"infer --receipt-url: {args.receipt_url}")
        if not args.release_url:
            if args.receipt_url.endswith(".json"):
                args.release_url = args.receipt_url[:-5]
                log.info(f"infer --release-url: {args.release_url}")
            else:
                err = "could not guess the --release-url, please provide it explicitly"
                raise ValueError(err)

        if not args.release_version:
            args.release_version = get_version(args.bundle)
            log.info(f"infer --release-version: {args.release_version}")
        if not args.release_version:
            err = "could not guess the --release-version, please provide it explicitly"
            raise ValueError(err)

        with ZipFile2(args.bundle, mode="a") as bundle:
            if not args.no_copy:
                copy_cosmofy(bundle, dry_run=args.dry_run)

            if not args.no_args:
                set_args(bundle, add_arg_prefix(get_args(bundle)), dry_run=args.dry_run)

            write_receipt(
                args.bundle,
                bundle,
                output=args.receipt,
                receipt_url=args.receipt_url,
                release_url=args.release_url,
                release_version=args.release_version,
                dry_run=args.dry_run,
            )
    except Exception as e:
        args.show_error(log, e)
        return 2
    return 0


cmd = Command("cosmofy.updater.add", Args, run)

if __name__ == "__main__":
    sys.exit(cmd.main())



#!/usr/bin/env python
# TODO When mypy learns that `zipfile.Path.at` exists, remove next line.
# mypy: disable-error-code="attr-defined"

# std
from __future__ import annotations
from dataclasses import dataclass
from datetime import datetime
from datetime import timedelta
from operator import attrgetter
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Iterator
from typing import Literal
from zipfile import Path as ZipPath
from zipfile import ZipInfo
import logging
import stat
import sys

# pkg
from cosmofy.args import common_args
from cosmofy.args import CommonArgs
from cosmofy.args import global_options
from cosmofy.baton import arg
from cosmofy.baton import Command
from cosmofy.zipfile2 import ZipFile2

from . import expand_glob
from . import shell_match

log = logging.getLogger(__name__)

usage = f"""\
List contents of a Cosmopolitan bundle.

Usage: cosmofy fs ls <BUNDLE> [OPTIONS] [FILE]...

Arguments:
{common_args}
  [FILE]...                 one or more file patterns to show

Filter options:
  -a, --all                 show entries whose name starts with `.`
  -B, --ignore-backups      hide entries whose name ends with `~`
      --hide <PATTERN>      hide matching entries, unless --all
  -I, --ignore <PATTERN>    hide matching entries, even with --all

Sort options:
  -r, --reverse             reverse the sort order
      --sort <MODE>         [choices: none, name, size, time, extension]
                            [default: name]

Output options:
  -l, --long                show permissions, size, and modified date
  -h, --human-readable      show sizes using powers of 1024 like 1K 2M 3G etc.
      --si                  show sizes using powers of 1000 (implies -h)

{global_options.replace("-h,", "   ")}
"""


@dataclass
class Args(CommonArgs):
    __doc__ = usage

    # positional
    file: list[str] = arg(list, positional=True)
    """Files to show information about."""

    # filter
    all: bool = arg(False, short="-a")
    """Whether to show entries that start with `.`"""

    ignore_backups: bool = arg(False, short="-B")
    """Hide entries that end with `~`"""

    hide: list[str] = arg(list)
    """Hide entries, **unless** with `--all`"""

    ignore: list[str] = arg(list, short="-I")
    """Hide entries, **even** with `--all`"""

    # sort
    reverse: bool = arg(False, short="-r")
    """Whether to reverse the sort."""

    sort: Literal["none", "name", "size", "time", "extension"] = arg("name")
    """How to sort the list."""

    # output
    long: bool = arg(False, short="-l")
    """Whether to use a long listing format."""

    human_readable: bool = arg(False, short="-h")
    """Whether to use human-readable sizes."""

    si: bool = arg(False)
    """Whether to use 1000 instead of 1024 for human-readable chunks."""


def human_size(n: float, si: bool = False) -> str:
    """Return a number of bytes in a human readable way."""
    chunk = 1000.0 if si else 1024.0
    units = (
        ("b", "k", "m", "g", "t", "p", "e", "z", "y")
        if si
        else ("B", "K", "M", "G", "T", "P", "E", "Z", "Y")
    )
    n = float(n)
    for u in units:
        if abs(n) < chunk or u in ["y", "Y"]:
            if u in ["b", "B"]:
                return f"{int(n)}{u}"
            s = f"{n:.1f}".rstrip("0").rstrip(".")
            return f"{s}{u}"
        n /= chunk
    return ""


def ls_time(dt: datetime, *, now: datetime | None = None) -> str:
    """Return a date/time representation similar to `ls`."""
    if now is None:
        now = datetime.now(dt.tzinfo)
    six_months = now - timedelta(days=182)

    mon = dt.strftime("%b")
    day = f"{dt.day:2d}"

    if six_months <= dt <= now and dt.year == now.year:
        hm = dt.strftime("%H:%M")
        return f"{mon} {day} {hm}"
    else:
        year = f"{dt.year:4d}"
        return f"{mon} {day}  {year}"


@dataclass
class Runner:
    bundle: ZipFile2
    args: Args

    def should_include(self, name: str) -> bool:
        """Return `True` if we should include this item."""
        args = self.args
        if any(shell_match(name, pat) for pat in args.ignore):  # always hide
            return False
        if args.all:
            return True
        elif any(shell_match(name, pat) for pat in args.hide):  # hide, if not --all
            return False
        elif name.startswith("."):  # hide .name, if not --all
            return False
        return True

    def get_zipinfo(self, path: ZipPath) -> ZipInfo:
        """Return `ZipInfo` for a path or create a synthetic record."""
        return self.bundle.NameToInfo.get(path.at, ZipInfo(path.at))

    def get_files(self) -> Iterator[ZipInfo]:
        """Iterate over files in a bundle, hiding entries as appropriate."""
        args = self.args
        root = ZipPath(self.bundle)
        seen: set[str] = set()
        names = self.bundle.namelist()

        for pat in args.file:
            for item in expand_glob(names, pat):
                path = root / item
                if item != "" and not path.exists():
                    raise FileNotFoundError(f"No such file or directory: {item}")

                if item != "":  # send the thing itself
                    if path.at not in seen and self.should_include(path.name):
                        seen.add(path.at)
                        yield self.get_zipinfo(path)

                if path.is_dir():
                    for sub in path.iterdir():
                        if sub.at not in seen and self.should_include(sub.name):
                            seen.add(sub.at)
                            yield self.get_zipinfo(sub)

    @staticmethod
    def get_extension(f: ZipInfo) -> tuple[str, str, str]:
        """Return string for extension sorting."""
        parts = f.filename.rstrip("/").split("/")
        if f.filename.endswith("/"):  # is dir
            return "/".join(parts[:-1]), "", parts[-1]

        name = parts[-1]
        stem, ext = name, ""
        if "." in name:
            stem, ext = name.rsplit(".", 1)
        return "/".join(parts[:-1]), ext, stem

    def sort(self, files: Iterable[ZipInfo]) -> Iterator[ZipInfo]:
        """Sort the selected files."""
        if self.args.sort == "none":
            yield from files
            return

        key: Callable[[ZipInfo], Any]
        reverse = self.args.reverse
        match self.args.sort:
            case "name":
                key = attrgetter("filename")  # default
            case "size":
                key = attrgetter("file_size")
                reverse = not reverse  # biggest first
            case "time":
                key = attrgetter("date_time")
                reverse = not reverse  # newest first
            case "extension":
                key = self.get_extension
            case _:  # pragma: no cover
                raise ValueError(f"Unknown option for `--sort`: {self.args.sort}")

        items = sorted(files, key=key, reverse=reverse)
        yield from items

    def format(self, f: ZipInfo) -> str:
        """Prepare info for printing."""
        args = self.args
        line: list[str] = []
        if args.long:
            line.append(f"{stat.filemode(f.external_attr >> 16)} ")

            if args.human_readable or args.si:
                line.append(f"{human_size(f.file_size, args.si):>10} ")
                # line.append(f"{human_size(f.compress_size, args.si):>10} ")
            else:
                line.append(f"{f.file_size:>10} ")
                # line.append(f"{f.compress_size:>10} ")

            line.append(f"{ls_time(datetime(*f.date_time))} ")

        line.append(f.filename)
        return "".join(line)

    def run(self) -> None:
        """List all files in a bundle."""
        for f in self.sort(self.get_files()):
            print(self.format(f))


def run(args: Args) -> int:
    "Entry point for `cosmofy fs ls`."
    args.setup_logger()
    try:
        assert args.ensure_bundle() and args.bundle
        if args.dry_run:
            log.info(f"{args.banner}list: {args.bundle}")
            return 0

        if args.ignore_backups:
            # https://www.gnu.org/software/coreutils/manual/html_node/Which-files-are-listed.html#index-_002dB
            args.ignore.extend(["*~", ".*~"])

        if not args.file:
            args.file.extend([""])  # root directory

        # good to go
        with ZipFile2(args.bundle) as bundle:
            Runner(bundle, args).run()
    except Exception as e:
        args.show_error(log, e)
        return 2
    return 0


cmd = Command("cosmofy.fs.ls", Args, run)

if __name__ == "__main__":
    sys.exit(cmd.main())



"""baton: Dataclass argument parser with subcommand handoff."""

# std
from __future__ import annotations
from dataclasses import dataclass
from dataclasses import field
from dataclasses import Field
from dataclasses import fields
from dataclasses import MISSING
from dataclasses import replace
from os import environ as ENV
from textwrap import dedent
from types import UnionType
from typing import Any
from typing import Callable
from typing import cast
from typing import get_args
from typing import get_origin
from typing import get_type_hints
from typing import Literal
from typing import TextIO
from typing import Union
import logging
import re
import sys

# TODO py3.10
if sys.version_info >= (3, 11):  # pragma: no cover
    from typing import assert_never
else:  # pragma: no cover

    def assert_never(arg: Any, /) -> None:
        value = repr(arg)
        raise AssertionError(f"Expected code to be unreachable, but got: {value}")


Action = Literal[
    "",
    "store",
    "store_bool",
    "store_true",
    "store_false",
    "append",
    "extend",
    "count",
]


@dataclass
class ArgMetadata:
    short: str = ""
    action: Action = ""  # infer, by default
    required: bool = False
    positional: bool = False
    env: str = ""


def arg(
    default: Any,
    *,
    short: str = "",
    env: str = "",
    positional: bool = False,
    required: bool = False,
    action: Action = "",
    **kwargs: Any,
) -> Any:
    """`dataclasses.field` wrapper for argument metadata."""
    if required and not positional:
        raise ValueError("optional arguments cannot be required")

    metadata = kwargs.pop("metadata", {})
    metadata["arg"] = ArgMetadata(
        short=short,
        action=action,
        required=required,
        positional=positional,
        env=env,
    )
    if "default_factory" not in kwargs:  # can't set both
        if default in (list, dict, set):
            kwargs["default_factory"] = default
        else:
            kwargs["default"] = default
    return field(metadata=metadata, **kwargs)


def get_choices(kind: Any) -> tuple[type, list[Any]] | None:
    """Extract choices from Literal type, if present."""
    if get_origin(kind) is Literal:
        choices = get_args(kind)
        # infer item type from first choice
        item_type = type(choices[0]) if choices else str
        return cast(type, item_type), list(choices)
    return None


def get_item_type(kind: type | UnionType) -> type:
    """Get the item type for list[X] -> X, else return kind."""
    origin = get_origin(kind)
    if origin in (list, Union, UnionType):
        args = get_args(kind)
        return cast(type, args[0]) if args else str
    return cast(type, kind)


def infer_action(kind: type, positional: bool = False) -> Action:
    """Infer action from type and default value."""
    origin = getattr(kind, "__origin__", None)
    if origin is list:
        return "extend"
    if kind is bool:
        return "store_bool" if positional else "store_true"
    return "store"


@dataclass
class Arg:
    long: str
    short: str
    field_name: str
    field_type: type
    item_type: type  # for containers, otherwise same as field_type
    choices: list[Any]
    action: Action
    required: bool
    positional: bool
    env: str

    @staticmethod
    def from_class(spec: type) -> list[Arg]:
        """Return list of arguments from a class."""
        hints = get_type_hints(spec)
        return [Arg.from_field(f, hints) for f in fields(spec)]

    @classmethod
    def from_field(cls, f: Field[Any], type_hints: dict[str, type]) -> Arg:
        """Return an argument constructed from a field."""
        kind = type_hints.get(f.name, str)
        if choices_info := get_choices(kind):
            item_type, choices = choices_info
        else:
            item_type = get_item_type(kind)
            choices = []

        meta: ArgMetadata | None = f.metadata.get("arg")
        if meta is not None:
            action = meta.action or infer_action(kind, meta.positional)
            short = meta.short
            required = meta.required
            positional = meta.positional
            env = meta.env
        else:  # need to infer
            action = infer_action(kind)
            short = ""
            required = f.default is MISSING
            positional = False
            env = ""

        return cls(
            long=f.name if positional else f"--{f.name.replace('_', '-')}",
            short=short,
            env=env,
            field_name=f.name,
            field_type=kind,
            item_type=item_type,
            choices=choices,
            action=action,
            positional=positional,
            required=required,
        )


def _init_context(cmd: Command, parent: Any | None) -> object:
    """Create context with appropriate defaults."""
    ctx = cmd.cls()

    for spec in cmd.args:
        if spec.env and spec.env in ENV:
            val = ENV[spec.env]
            if spec.action in ("store_true", "store_false"):
                _do_action(ctx, replace(spec, action="store_bool"), val)
            else:
                _do_action(ctx, spec, val)
    # env defaults set

    if parent:
        parent_fields = {f.name for f in fields(parent)}
        for f in fields(cmd.cls):
            if f.name in parent_fields:
                setattr(ctx, f.name, getattr(parent, f.name))
    # parent parsed values set

    return ctx


def _pop_values(argv: list[str], stop: set[str] | None = None) -> list[str]:
    """Pop values from argv until a flag or stop condition."""
    stop = stop or set()
    vals = []
    while argv and not argv[0].startswith("-") and argv[0] not in stop:
        vals.append(argv.pop(0))
    return vals


def _do_action(ctx: object, spec: Arg, val: str | list[str] = "") -> None:
    """Apply a parsed value to the context object."""
    name = spec.field_name
    kind = spec.item_type
    match spec.action:
        case "store":
            if spec.choices and val not in spec.choices:
                err = f"invalid value '{val}' for {spec.long}"
                err += f"\n  [choices: {', '.join(spec.choices)}]"
                raise ValueError(err)
            setattr(ctx, name, kind(val))
        case "store_bool":
            assert isinstance(val, str)
            setattr(ctx, name, val.lower() in ("1", "true", "yes", "on", "y", "t"))
        case "store_true":
            setattr(ctx, name, True)
        case "store_false":
            setattr(ctx, name, False)
        case "append":
            getattr(ctx, name).append(kind(val))
        case "extend":
            getattr(ctx, name).extend(kind(v) for v in val)
        case "count":
            setattr(ctx, name, getattr(ctx, name) + 1)
        case _:  # pragma: no cover
            assert_never(spec.action)


def _parse_optional(ctx: object, spec: Arg, argv: list[str]) -> None:
    """Parse an optional argument, consuming values from argv."""
    match spec.action:
        case "count" | "store_true" | "store_false":
            _do_action(ctx, spec, "")
        case "store" | "store_bool" | "append":
            if not argv:
                err = f"a value is required for {spec.long}, but none was supplied"
                if spec.choices:
                    err += f"\n  [choices: {', '.join(spec.choices)}]"
                raise ValueError(err)
            _do_action(ctx, spec, argv.pop(0))
        case "extend":
            vals = _pop_values(argv)
            if not vals:
                err = f"a value is required for {spec.long}, but none was supplied"
                if spec.choices:
                    err += f"\n  [choices: {', '.join(spec.choices)}]"
                raise ValueError(err)
            _do_action(ctx, spec, vals)
        case _:  # pragma: no cover
            assert_never(spec.action)


def _parse(
    cmd: Command, argv: list[str], parent: Any | None = None
) -> tuple[Any, Command | None, list[str]]:
    """Parse `argv` into `cmd.cls` instance.

    Returns: (parsed_args, subcommand_name | None, remaining_argv)
    """
    ctx = _init_context(cmd, parent)
    aliases = cmd.aliases
    optionals = cmd.optionals
    positionals = list(cmd.positionals)
    subcommand: Command | None = None

    while argv:
        val = argv.pop(0)

        # Handle --foo=bar
        if "=" in val and val.startswith("-"):
            val, rest = val.split("=", 1)
            argv.insert(0, rest)

        # Rest are positional
        if val == "--":
            if positionals and positionals[0].action == "extend":
                _do_action(ctx, positionals.pop(0), argv.copy())
            elif positionals:
                for v in argv:
                    if not positionals:
                        raise ValueError(f"unexpected argument: '{v}'")
                    _do_action(ctx, positionals.pop(0), v)
            elif argv:
                raise ValueError(f"unexpected argument: '{argv[0]}'")
            argv = []
            break

        # Expand -vvv -> --verbose --verbose --verbose
        if val.startswith("-") and not val.startswith("--") and " " not in val:
            expanded = []
            for c in val[1:]:
                if alias := aliases.get(f"-{c}"):
                    expanded.append(alias)
                else:
                    raise ValueError(f"unknown option: '-{c}'")
            argv = expanded + argv
            continue

        # Optional
        if val.startswith("--") and " " not in val:
            spec = optionals.get(val)
            if not spec:
                raise ValueError(f"unknown option: '{val}'")
            _parse_optional(ctx, spec, argv)
            continue

        # Positional
        if not positionals:
            raise ValueError(f"unexpected argument: '{val}'")

        spec = positionals.pop(0)
        # TODO: Mark the positional that is supposed to receive the subcommand name.
        if spec.long == "command" and val not in cmd.subcommands:
            raise ValueError(f"unknown subcommand name: '{val}'")

        if spec.action == "extend":
            vals = [val] + _pop_values(argv, set(cmd.subcommands.keys()))
            _do_action(ctx, spec, vals)
        else:
            _do_action(ctx, spec, val)

        # Subcommand
        if val in cmd.subcommands:
            subcommand = cmd.subcommands[val]
            break

    # Check required
    missing = [f"<{s.field_name.upper()}>" for s in positionals if s.required]
    if missing and not getattr(ctx, "help", None):
        s = "s" if len(missing) > 1 else ""
        raise ValueError(f"missing required argument{s}: {', '.join(missing)}")

    return ctx, subcommand, argv


def parse(cls: Any, argv: list[str]) -> Any:
    """Parse `argv` into an instance of `cls` without any subcommands."""
    cmd = Command(cls.__name__, cls, lambda _: 0)
    ctx, _, _ = _parse(cmd, argv)
    return ctx


@dataclass
class Command:
    name: str
    cls: type  # The dataclass for args
    run: Callable[[Any], int]  # Takes parsed args, returns exit code
    usage: str = ""
    subcommands: dict[str, Command] = field(default_factory=dict)

    _args: list[Arg] = field(default_factory=list, init=False, repr=False)

    def __post_init__(self) -> None:
        """Make some adjustments after the dataclass is built."""
        self._args = Arg.from_class(self.cls)
        if not self.usage:
            self.usage = dedent(self.cls.__doc__ or "")

    @property
    def args(self) -> list[Arg]:
        """Return the current args."""
        return self._args

    @property
    def aliases(self) -> dict[str, str]:
        """Return argument aliases mapped to their long counter-parts."""
        return {a.short: a.long for a in self._args if a.short}

    @property
    def optionals(self) -> dict[str, Arg]:
        """Return optional arguments."""
        return {a.long: a for a in self._args if not a.positional}

    @property
    def positionals(self) -> list[Arg]:
        """Return list of positional arguments."""
        return [a for a in self._args if a.positional]

    def show_usage(self, *, short: bool = False, color: COLOR_MODE = "auto") -> None:
        """Display the usage for this command."""
        usage = self.usage
        if short:
            beg = usage.find("Usage:")
            end: int | None = usage.find("\n\n", beg)
            if end == -1:
                end = None  # go to the end
            usage = f"\n{usage[beg:end]}\n\nFor more information, try --help."
        else:
            usage = self.usage.strip()
        if use_color(color, sys.stdout):
            usage = decorate(usage)
        print(render_tags(usage, color=color))

    parse = _parse

    def main(
        self: Command, argv: list[str] | None = None, parent: object | None = None
    ) -> int:
        """Parse and execute the command."""
        log = logging.getLogger(self.name)
        if argv is None:
            argv = sys.argv[1:]

        try:
            args, sub, remaining = _parse(self, argv, parent)
            if getattr(args, "help", False):
                self.show_usage(color=getattr(args, "color", "auto"))
                return 0

            if sub:
                return sub.main(remaining, parent=args)
            return self.run(args)
        except ValueError as e:
            log.error(e)
            self.show_usage(short=True)
            return 1


COLOR_MODE = Literal["auto", "always", "never"]
COLOR_CODE: dict[str, str] = {
    "reset": "0",
    "bold": "1",
    "dim": "2",
    "italic": "3",
    "underline": "4",
    "black": "30",
    "red": "31",
    "green": "32",
    "yellow": "33",
    "blue": "34",
    "magenta": "35",
    "cyan": "36",
    "white": "37",
}

DEFAULT_THEME = {
    "heading": "bold green",  # Heading:
    "command": "bold cyan",  # command
    "argument": "cyan",  # <ARGUMENT>
    "option": "cyan",  # [OPTION]
    "flag": "bold cyan",  # --flag
    "repeats": "cyan",  # ...
    "tick": "bold white",  # `tick`
    "string": "yellow",  # 'yellow'
    "choice": "green",  # [choices: a, b, c]
    "default": "yellow",  # [default: value]
    "env": "white",  # [env: NAME=value]
    "tip": "green",  # tip:
}


class ColorFormatter(logging.Formatter):
    """Log formatter that renders [tag]...[/] markup to ANSI colors."""

    color: COLOR_MODE
    _stream: TextIO | None

    def __init__(
        self,
        fmt: str | None = None,
        datefmt: str | None = None,
        style: Literal["%", "{", "$"] = "%",
        *,
        color: COLOR_MODE = "auto",
    ):
        """Construct a new `ColorFormatter`."""
        super().__init__(fmt, datefmt, style)
        self.color = color
        self._stream = None  # check at format time

    def format(self, record: logging.LogRecord) -> str:
        """Format the log record."""
        match record.levelname:
            case "ERROR":
                level = "[bold red]error[/]"
            case "WARNING":
                level = "[bold yellow]warning[/]"
            case "INFO":
                level = "[bold cyan]info[/]"
            case "DEBUG":
                level = "[dim]debug[/]"
            case _:
                level = record.levelname.lower()

        record.levelname = level
        message = super().format(record)
        stream = self._stream or sys.stderr
        if self.color in ("auto", "always"):
            message = decorate(message)
        return render_tags(message, color=self.color, file=stream)

    def set_stream(self, stream: TextIO) -> None:
        """Set the stream for TTY detection (called by handler)."""
        self._stream = stream


class ColorHandler(logging.StreamHandler):  # type: ignore
    """StreamHandler that automatically configures ColorFormatter's stream.

    Usage:
        import logging
        from baton import ColorHandler

        handler = ColorHandler()
        handler.setFormatter(ColorFormatter("[dim]%(asctime)s[/] %(message)s"))
        logging.getLogger().addHandler(handler)
    """

    def setFormatter(self, fmt: logging.Formatter | None) -> None:
        """Set the formatter."""
        super().setFormatter(fmt)
        if isinstance(fmt, ColorFormatter):
            fmt.set_stream(self.stream)


def use_color(color: COLOR_MODE, file: Any) -> bool:
    """Return `True` if we should output ANSI color codes."""
    # See: https://no-color.org/
    # See: https://bixense.com/clicolors/

    # cli
    if color == "never":
        return False
    if color == "always":
        return True

    # env
    if ENV.get("NO_COLOR") or ENV.get("CLICOLOR") in ["0", "false"]:
        return False
    if ENV.get("FORCE_COLOR") or ENV.get("CLICOLOR_FORCE"):
        return True

    # auto
    if not hasattr(file, "isatty") or not file.isatty():
        return False
    return ENV.get("TERM", "") != "dumb"


def render_tags(
    text: str,
    *,
    color: COLOR_MODE = "auto",
    file: Any | None = None,
    theme: dict[str, str] | None = None,
) -> str:
    """Render [tag]...[/] markup to ANSI (or strip if no color)."""
    if theme is None:
        theme = DEFAULT_THEME

    def replace_tag(m: re.Match[str]) -> str:
        """Return the string with the tag replaced."""
        tag = m.group(1).lower()
        if tag == "/":
            tag = "reset"
        tag = theme.get(tag, tag)
        parts = tag.split()
        codes = [COLOR_CODE[p] for p in parts if p in COLOR_CODE]
        if codes:
            return f'\033[{";".join(codes)}m'
        return m.group(0)  # Unknown tag, leave as-is

    if use_color(color, sys.stdout if file is None else file):
        return re.sub(r"\[([^[\]]+)\]", replace_tag, text)
    else:
        return re.sub(r"\[/?[a-z ]*\]", "", text)


def decorate(text: str) -> str:
    """Apply theme markings."""
    result = text

    # Usage: command
    # starts with usage and everything up to the first non lowercase word
    result = re.sub(r"Usage: ([a-z ]+)", r"Usage: [command]\1[/]", result)

    # Heading:
    # start of a line, starts with a capital letter ends with a colon
    result = re.sub(r"\n([A-Z][^\n:]+:)", r"\n[heading]\1[/]", result)

    # command
    # two spaces before and after, all lowercase, can have dashes
    result = re.sub(r"  ([a-z][-_a-z]+)  ", r"  [command]\1[/]  ", result)

    # [OPTION], [OPTION...]
    result = re.sub(r"\[([<A-Z_>]+)\]", r"[option][\1][/]", result)

    # <ARGUMENT>
    result = re.sub(r"<([A-Z_]+)>", r"[argument]<\1>[/]", result)

    # `--flag`, --flag
    result = re.sub(r"`(-[-A-Za-z]+)`", r"[flag]\1[/]", result)
    result = re.sub(r"(\s+)(-[-A-Za-z]+)", r"\1[flag]\2[/]", result)

    # `tick`
    result = re.sub(r"`([^`]+)`", r"[tick]\1[/]", result)

    # 'string'
    result = re.sub(r"'([^']+)'", r"'[string]\1[/]'", result)

    # repeats
    result = re.sub(r"([a-z\]>])(\.\.\.)", r"\1[repeats]\2[/]", result)

    # tip:
    result = re.sub(r"  tip:", r"  [tip]tip:[/]", result)

    # [default: value]
    result = re.sub(r"\[default: ([^\]]+)\]", r"[default: [default]\1[/]]", result)

    # [env: NAME=value]
    result = re.sub(
        r"\[env: ([^=]+)=([^\]]*)\]", r"[env: [env]\1[/]=[default]\2[/]]", result
    )

    # [choices: a, b, c]
    result = re.sub(
        r"\[(choices|possible values):\s*([^\]]+)\]",
        lambda m: f"[{m.group(1)}: "
        + re.sub(
            r"\s*,\s*",
            ", ",
            re.sub(r"\s*([^,\]]+)\s*", r"[choice]\1[/]", m.group(2)).strip(),
        )
        + "]",
        result,
    )

    return result



"""A Cosmofy Receipt is metadata for self-updating programs."""

# std
from __future__ import annotations
from datetime import datetime
from datetime import timezone
from pathlib import Path
from typing import Callable
from urllib.request import urlopen
import dataclasses
import hashlib
import json
import logging
import re
import shlex
import subprocess

# pkg
from . import DEFAULT_HASH
from . import COSMOFY_TIMEOUT

log = logging.getLogger(__name__)

Checker = Callable[[str], bool]
"""Function that takes a `str` and returns a `bool` if it is ok."""

RECEIPT_SCHEMA = (
    "https://raw.githubusercontent.com/metaist/cosmofy/0.1.0/cosmofy.schema.json"
)
"""URI of Cosmofy Receipt Schema."""

RECEIPT_KIND = ("embedded", "published")
"""Valid values for `Receipt.kind`."""

RECEIPT_DATE = re.compile(r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z$")
"""Regex to validate `Receipt.date`."""

RECEIPT_ALGO = re.compile(r"^[a-z0-9-_]+$")
"""Regex to validate `Receipt.algo`."""

RECEIPT_HASH = re.compile(r"^[a-f0-9]+$")
"""Regex to validate `Receipt.hash`."""

RE_VERSION = re.compile(r"\d+\.\d+\.\d+(-[\da-zA-Z-.]+)?(\+[\da-zA-Z-.]+)?")
"""Regex for a semver-like version string."""


def datestr(date: datetime) -> str:
    """Return an ISO-8601 formatted date string.

    >>> datestr(datetime(2000, 1, 1, tzinfo=timezone.utc))
    '2000-01-01T00:00:00Z'
    """
    return date.astimezone(timezone.utc).isoformat()[:19] + "Z"


def get_version(path: Path, default: str = "") -> str:
    """Return `path` version or `default` if it doesn't exist.

    >>> get_version(Path("fake"), "0.0.0")
    '0.0.0'
    """
    version = default
    if path.is_file():
        try:
            # NOTE: We need to use `shell=True` because the path is a Cosmo APE.
            # Otherwise we get: `OSError: [Errno 8] Exec format error`
            cmd = f"{shlex.quote(str(path.resolve()))} --version"
            out: str = subprocess.check_output(cmd, shell=True, text=True)
            if match := RE_VERSION.search(out):
                version = match.group()
        except subprocess.CalledProcessError as e:  # we can't get the version
            log.exception(e)
            return default
    return version


@dataclasses.dataclass
class Receipt:
    """Asset metadata."""

    schema: str = RECEIPT_SCHEMA
    """Receipt schema."""

    kind: str = RECEIPT_KIND[0]
    """Whether this receipt is full (published) or partial (embedded)."""

    date: str = datestr(datetime.now())
    """UTC date/time of this receipt."""

    algo: str = DEFAULT_HASH
    """Hashing algorithm."""

    hash: str = ""
    """Asset hash."""

    receipt_url: str = ""
    """Asset receipt URL."""

    release_url: str = ""
    """Asset download URL."""

    version: str = ""
    """Asset version."""

    def is_newer(self, other: Receipt) -> bool:
        """Return `True` if this receipt is newer than `other`."""
        return self.date > other.date

    def __str__(self) -> str:
        """Return `json`-encoded string."""
        return json.dumps(self.asdict())

    def asdict(self) -> dict[str, str]:
        """Return `dict` representation of the receipt."""
        return {
            "$schema": self.schema,
            "kind": self.kind,
            "date": self.date,
            "algo": self.algo,
            "hash": self.hash,  # maybe empty
            "receipt_url": self.receipt_url,
            "release_url": self.release_url,
            "version": self.version,  # maybe empty
        }

    def is_valid(self) -> bool:
        """Return `True` if there are no issues with the receipt."""
        issues = Receipt.find_issues(self.asdict())
        return not sum((v for v in issues.values()), [])

    @staticmethod
    def find_issues(data: dict[str, str]) -> dict[str, list[str]]:
        """Return field names by issue that occurred during validation."""
        issues: dict[str, list[str]] = {"missing": [], "unknown": [], "malformed": []}
        rules: dict[str, Checker] = {
            "$schema": lambda v: v == RECEIPT_SCHEMA,
            "kind": lambda v: v in RECEIPT_KIND,
            "date": lambda v: bool(RECEIPT_DATE.match(v)),
            "algo": lambda v: bool(RECEIPT_ALGO.match(v)),
            "hash": lambda v: bool(RECEIPT_HASH.match(v)),
            "receipt_url": lambda v: bool(v.strip()),
            "release_url": lambda v: bool(v.strip()),
            "version": lambda v: bool(v.strip()),
        }
        embedded: dict[str, Checker] = {
            "hash": lambda v: isinstance(v, str),
            "version": lambda v: isinstance(v, str),
        }

        kind = data.get("kind", "embedded")
        issues["unknown"] = [name for name in data if name not in rules]
        for name, rule in rules.items():
            if name not in data:
                issues["missing"].append(name)
                continue
            if name in embedded and kind == "embedded":
                rule = embedded[name]
            if not rule(data[name]):
                issues["malformed"].append(name)
        return issues

    def update(self, **values: str) -> Receipt:
        """Update this receipt with several values."""
        for name, value in values.items():
            setattr(self, name, value)
        return self

    def update_from(self, other: Receipt, *names: str, **values: str) -> Receipt:
        """Update this receipt from another receipt."""
        for name in names:
            setattr(self, name, getattr(other, name))
        return self.update(**values)

    @staticmethod
    def from_dict(data: dict[str, str]) -> Receipt:
        """Return receipt from a `dict`."""
        issues = Receipt.find_issues(data)
        if sum((v for v in issues.values()), []):
            raise ValueError("Invalid receipt", issues)

        schema = data["$schema"]
        _data = {k: v for k, v in data.items() if k != "$schema"}
        return Receipt(schema=schema, **_data)

    @staticmethod
    def from_url(url: str, timeout: int = COSMOFY_TIMEOUT) -> Receipt:
        """Return a Receipt from a URL."""
        with urlopen(url, timeout=timeout) as response:
            return Receipt.from_dict(json.load(response))

    @staticmethod
    def from_path(path: Path, version: str = "", algo: str = DEFAULT_HASH) -> Receipt:
        """Return hash and version for a `path`."""
        digest = hashlib.new(algo, path.read_bytes()).hexdigest()
        if not version:
            version = get_version(path)
        return Receipt(algo=algo, hash=digest, version=version)



"""Download files."""

# std
from datetime import datetime
from datetime import timezone
from email.utils import parsedate_to_datetime
from http.client import HTTPResponse
from pathlib import Path
from typing import Iterator
from urllib.error import HTTPError
from urllib.parse import urlparse
from urllib.request import Request
from urllib.request import urlopen
import hashlib
import logging
import os
import platform
import stat
import sys
import tempfile

# pkg
from . import DEFAULT_HASH
from . import COSMOFY_TIMEOUT

log = logging.getLogger(__name__)

CHUNK_SIZE = 65536
"""Default chunk size."""


def validate_url(url: str, allow_http: bool = False) -> bool:
    """Validate URL scheme is https (or http if explicitly allowed)."""
    parsed = urlparse(url)
    allowed = ("https",) if not allow_http else ("https", "http")
    if parsed.scheme not in allowed:
        raise ValueError(f"URL must use HTTPS, got: {parsed.scheme}://")
    return True


def move_executable(src: Path, dest: Path) -> Path:
    """Set the executable bit and move a file."""
    mode = src.stat().st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
    src.chmod(mode)
    dest.parent.mkdir(parents=True, exist_ok=True)
    if dest == Path(sys.executable) and platform.system() == "Windows":
        # can't overwrite running executable
        old = dest.with_suffix(dest.suffix + ".old")
        if old.exists():  # from previous update
            old.unlink()
        if dest.exists():  # rename running exe
            os.rename(dest, old)
        # Note: .old file will be cleaned up on next update

    os.replace(src, dest)  # atomic move
    log.debug(f"move: {src} to {dest}")
    return dest


def cleanup_old_executable(path: Path) -> None:
    """Remove leftover .old file from previous Windows update."""
    old = path.with_suffix(path.suffix + ".old")
    if old.exists():
        try:
            old.unlink()
            log.debug(f"Cleaned up old executable: {old}")
        except OSError:
            pass  # Still in use or permission denied, ignore


def progress(response: HTTPResponse, prefix: str = "Downloading: ") -> Iterator[bytes]:
    """Display progress information."""
    header = response.getheader("Content-Length") or "0"
    total = int(header.strip())
    done = 0
    while chunk := response.read(CHUNK_SIZE):
        done += len(chunk)
        if total > 0:
            percent = done / total * 100
            print(f"\r{prefix}{percent:.2f}%", end="", flush=True)
        else:
            print(f"\r{prefix}{done} bytes", end="", flush=True)
        yield chunk
    print("")


def download(url: str, path: Path, timeout: int = COSMOFY_TIMEOUT) -> Path:
    """Download `url` to path."""
    validate_url(url)
    log.info(f"download: {url} to {path}")

    path.parent.mkdir(parents=True, exist_ok=True)
    tmp = path.with_suffix(path.suffix + ".tmp")
    try:
        with urlopen(url, timeout=timeout) as response, tmp.open("wb") as output:
            for chunk in progress(response):
                output.write(chunk)
        os.replace(tmp, path)
    except:  # handle partial downloads
        tmp.unlink(missing_ok=True)
        raise
    return path


def download_if_newer(url: str, path: Path, timeout: int = COSMOFY_TIMEOUT) -> Path:
    """Download `url` to `path` if `url` is newer."""
    validate_url(url)
    if not path.exists():
        return download(url, path)

    with urlopen(Request(url, method="HEAD"), timeout=timeout) as response:
        last_modified = response.headers.get("Last-Modified")
        if not last_modified:
            log.debug("no `Last-Modified` header; re-downloading")
            return download(url, path)

    try:
        local = datetime.fromtimestamp(path.stat().st_mtime, tz=timezone.utc)
        remote = parsedate_to_datetime(last_modified)
        if remote > local:
            return download(url, path)
        return path  # cached version is current
    except (TypeError, ValueError):
        log.debug(f"could not parse `Last-Modified`: {last_modified}; re-downloading")
        return download(url, path)


def download_and_hash(
    url: str, path: Path, algo: str = DEFAULT_HASH, timeout: int = COSMOFY_TIMEOUT
) -> str:
    """Download `url` to `path` and return the hash."""
    validate_url(url)
    log.info(f"download: {url} to {path}")
    digest = hashlib.new(algo)

    path.parent.mkdir(parents=True, exist_ok=True)
    tmp = path.with_suffix(path.suffix + ".tmp")
    try:
        with urlopen(url, timeout=timeout) as response, tmp.open("wb") as output:
            for chunk in progress(response):
                digest.update(chunk)
                output.write(chunk)
    except:  # handle partial download
        tmp.unlink(missing_ok=True)
        raise
    return digest.hexdigest()


def download_release(
    url: str, path: Path, expected: str, algo: str = DEFAULT_HASH
) -> Path | None:
    """Download release from `url` checking the hash along the way."""
    validate_url(url)
    log.info(f"download {url} to {path}")
    with tempfile.NamedTemporaryFile(delete=False) as out:
        temp = Path(out.name)
        try:
            received = download_and_hash(url, temp, algo)
        except HTTPError as e:
            log.error(f"{e}: {url}")
            return None

        if received != expected:
            log.error(f"hash mismatch: expected={expected}, received={received}")
            temp.unlink(missing_ok=True)
            return None

    log.debug(f"overwriting: {path}")
    return move_executable(temp, path)



#!/usr/bin/env python
"""Cosmofy self-updater."""

# std
from __future__ import annotations
from os import environ as ENV
from pathlib import Path
import logging
import sys
import zipfile

# pkg
from cosmofy import __pubdate__
from cosmofy import __version__

from .add import RECEIPT_URL
from .add import RELEASE_URL
from .check import check
from .downloader import cleanup_old_executable
from .downloader import download_release
from .pythonoid import run_python

log_normal = "%(levelname)s: %(message)s"
log_debug = "%(name)s.%(funcName)s: %(levelname)s: %(message)s"
log = logging.getLogger(__name__)


usage = f"""\
This program is bundled into Cosmopolitan Python apps
to give them the ability to update themselves.
See: https://github.com/metaist/cosmofy

Usage: <bundle> --self-update [--help] [--version] [--debug]

Options:
  --self-update     Run this updater instead of <bundle>
  -h, --help        Show this message and exit.
  --version         Show updater version and exit.
  --debug           Show debug messages.

  [env: RECEIPT_URL={RECEIPT_URL}]
  Override the embedded URL for downloading update metadata.

  [env: RELEASE_URL={RELEASE_URL}]
  Override the published URL for downloading the update.
"""


def self_update(path: Path) -> int:
    """Run the self-updater."""
    with zipfile.ZipFile(path, "r") as bundle:
        is_newer, local, remote = check(bundle, receipt_url=RECEIPT_URL, dry_run=False)
        if not is_newer:
            return 0

    url = ENV.get("RELEASE_URL", remote.release_url)
    dest = download_release(url, path, remote.hash, remote.algo)
    if not dest:
        return 1

    log.info(f"Updated to {remote.version}")
    return 0


def main(argv: list[str] | None = None) -> int:
    """Main entry point for self-updater."""
    cleanup_old_executable(Path(sys.executable))

    args = argv or sys.argv[1:]
    if "--self-update" in args:
        if "-h" in args or "--help" in args:
            print(usage, end="")
            return 0
        if "--version" in args:
            print(f"cosmofy.updater {__version__} ({__pubdate__})")
            return 0
        level = logging.DEBUG if "--debug" in args else logging.INFO
        logging.basicConfig(level=level, format=log_normal)
        return self_update(Path(sys.executable))
    return run_python(args)


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())



"""Command-line arguments."""

# std
from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
from zipfile import is_zipfile
import logging

# pkg
from .baton import arg
from .baton import COLOR_MODE
from .baton import ColorFormatter

DRY_RUN = "[`DRY RUN`] "
"""Banner to display when using `--dry-run`."""


def get_banner(dry_run: bool = False) -> str:
    """Return `DRY_RUN` banner if `dry_run` is `True`."""
    return DRY_RUN if dry_run else ""


log_normal = "%(levelname)s: %(message)s"
log_debug = "[dim]%(name)s.%(funcName)s: [/]%(levelname)s: %(message)s"
log_verbose = "[dim]%(filename)s:%(lineno)s %(funcName)s: [/]%(levelname)s: %(message)s"

log = logging.getLogger(__name__)

global_options = """\
Global options:
  -h, --help                show this help message
  -q, --quiet...            show quiet output
  -v, --verbose...          show verbose output
      --dry-run             do not make any filesystem changes
      --color <COLOR>       control output color
                            [default: auto][choices: auto, always, never]
                            (auto checks `NO_COLOR`, `FORCE_COLOR`, `CLICOLOR`,
                            `CLICOLOR_FORCE` and tty support)
""".rstrip()


@dataclass
class GlobalArgs:
    help: bool = arg(False, short="-h")
    quiet: int = arg(0, short="-q", action="count")
    verbose: int = arg(0, short="-v", action="count")
    dry_run: bool = arg(False)
    color: COLOR_MODE = arg(default="auto")

    @property
    def verbosity(self) -> int:
        """Return how verbose vs quiet we should be."""
        return self.verbose - self.quiet

    @property
    def for_real(self) -> bool:
        """Opposite of `dry_run`."""
        return not self.dry_run

    @for_real.setter
    def for_real(self, value: bool) -> None:
        """Set `dry_run`."""
        self.dry_run = not value

    @property
    def banner(self) -> str:
        """Return dry run banner, if required."""
        return get_banner(self.dry_run)

    def setup_logger(self) -> None:
        """Ensure logging is configured properly."""
        level = self.verbosity
        if level < 0:
            logging.disable(logging.CRITICAL)
        elif level > 0:
            root_logger = logging.getLogger()
            root_logger.setLevel(logging.DEBUG)
            fmt = log_verbose if level > 1 else log_debug
            formatter = ColorFormatter(fmt)
            for handler in root_logger.handlers:
                handler.setFormatter(formatter)
        log.debug(self)

    def show_error(self, log: logging.Logger, e: Exception) -> None:
        "Display an error."
        if self.verbosity > 1:
            log.exception(e)
        else:
            log.error(e)


common_args = """\
  <BUNDLE>                  Cosmopolitan file bundle
""".rstrip()


@dataclass
class CommonArgs(GlobalArgs):
    bundle: Path | None = arg(None, positional=True, required=True)

    def ensure_bundle(self) -> bool:
        """Resolve `bundle` parameter and make sure it exists (in non-dry run)."""
        if self.bundle:
            self.bundle = self.bundle.resolve()

        if self.for_real and (
            self.bundle is None
            or not self.bundle.exists()
            or not is_zipfile(self.bundle)
        ):
            raise FileNotFoundError(f"could not find Cosmopolitan file: {self.bundle}")

        return True



#!/usr/bin/env python

# std
from __future__ import annotations
from dataclasses import dataclass
from zipfile import Path as ZipPath
import logging
import sys

# pkg
from cosmofy.args import common_args
from cosmofy.args import CommonArgs
from cosmofy.args import get_banner
from cosmofy.args import global_options
from cosmofy.baton import arg
from cosmofy.baton import Command
from cosmofy.zipfile2 import ZipFile2

from . import expand_glob

log = logging.getLogger(__name__)

usage = f"""\
Remove files from a Cosmopolitan bundle.

Usage: cosmofy fs rm <BUNDLE> [OPTIONS] <FILE>...

Arguments:
{common_args}
  <FILE>...                 files to remove

  tip: Use `--` to separate options from filenames that start with `-`
  Example: cosmofy fs rm bundle.zip -- -weird-filename.txt

Options:
  -f, --force               ignore nonexistent files
  -r, --recursive           recursively remove directories

{global_options}
"""


@dataclass
class Args(CommonArgs):
    __doc__ = usage
    file: list[str] = arg(list, positional=True, required=True, action="extend")
    force: bool = arg(False, short="-f")
    recursive: bool = arg(False, short="-r")  # TODO: support -R


def remove_path(
    bundle: ZipFile2,
    name: str,
    *,
    force: bool = False,
    recursive: bool = False,
    # global
    dry_run: bool = False,
    level: int = logging.INFO,
) -> None:
    """Remove `path` from `bundle`."""
    banner = get_banner(dry_run)
    for_real = not dry_run

    path = ZipPath(bundle.filename or "", name)
    if not path.exists():
        if force:
            return
        raise FileNotFoundError(f"cannot find {name}")

    if path.is_file():
        if for_real:
            bundle.remove(name)
        log.log(level, f"{banner}removed: {name}")
    elif path.is_dir():
        if not recursive:
            err = f"cannot remove directory {name}"
            err += "\n  tip: use --recursive"
            raise Exception(err)
        for item in path.iterdir():
            remove_path(
                bundle,
                item.at,  # type: ignore
                force=force,
                recursive=recursive,
                dry_run=dry_run,
            )
        if name in bundle.NameToInfo:  # dir actually has an entry
            if for_real:
                bundle.remove(name)
            log.log(level, f"{banner}removed: {name}")


def run(args: Args) -> int:
    """Entry point for `cosmofy fs rm`."""
    args.setup_logger()
    try:
        assert args.ensure_bundle() and args.bundle
        # good to go

        with ZipFile2(args.bundle, mode="a") as bundle:
            names = bundle.namelist()
            for pat in args.file:
                for name in expand_glob(names, pat):
                    remove_path(
                        bundle,
                        name,
                        force=args.force,
                        recursive=args.recursive,
                        dry_run=args.dry_run,
                    )
    except Exception as e:
        args.show_error(log, e)
        return 2
    return 0


cmd = Command("cosmofy.fs.rm", Args, run)

if __name__ == "__main__":
    sys.exit(cmd.main())



#!/usr/bin/env python

# std
from __future__ import annotations
from dataclasses import dataclass
import logging
import sys

# pkg
from cosmofy.args import common_args
from cosmofy.args import CommonArgs
from cosmofy.args import global_options
from cosmofy.baton import arg
from cosmofy.baton import Command
from cosmofy.fs import expand_glob
from cosmofy.fs.args import get_args
from cosmofy.fs.args import set_args
from cosmofy.fs.rm import remove_path
from cosmofy.zipfile2 import ZipFile2

from .add import COSMOFY_NO_ARGS
from .add import PATH_COSMOFY

log = logging.getLogger(__name__)

ARGS_PREFIXES: list[str] = [
    "-m cosmofy.updater.run",  # 0.2.0
    "-m cosmofy.updater",  # 0.1.0
]

usage = f"""\
EXPERIMENTAL: Remove cosmofy self-updater from a cosmofy bundle.

Usage: cosmofy updater remove <BUNDLE> [OPTIONS]

Arguments:
{common_args}

Process options:
      --no-args             skip setting `.args`
                            [env: COSMOFY_NO_ARGS={COSMOFY_NO_ARGS}]

{global_options}
"""


@dataclass
class Args(CommonArgs):
    __doc__ = usage
    no_args: bool = arg(False, env="COSMOFY_NO_ARGS")


def remove_arg_prefix(python_args: str) -> str:
    """Remove cosmofy updater prefix, if any."""
    for prefix in ARGS_PREFIXES:
        if python_args.startswith(prefix):
            return python_args[len(prefix) :].strip()
    return python_args


def run(args: Args) -> int:
    """Entry point for `cosmofy updater remove`."""
    args.setup_logger()
    log.warning("EXPERIMENTAL: `cosmofy updater` operations are experimental")
    try:
        assert args.ensure_bundle() and args.bundle

        with ZipFile2(args.bundle, mode="a") as bundle:
            for name in expand_glob(bundle.namelist(), PATH_COSMOFY + "**"):
                remove_path(
                    bundle,
                    name,
                    force=True,
                    recursive=True,
                    dry_run=args.dry_run,
                )

            if not args.no_args:
                set_args(
                    bundle,
                    remove_arg_prefix(get_args(bundle)),
                    dry_run=args.dry_run,
                )
    except Exception as e:
        args.show_error(log, e)
        return 2
    return 0


cmd = Command("cosmofy.updater.remove", Args, run)

if __name__ == "__main__":
    sys.exit(cmd.main())



# std
from os import environ as ENV

DEFAULT_HASH = "sha256"
"""Default hashing algorithm."""

DEFAULT_TIMEOUT = 30  # seconds
"""Default network timeout."""

COSMOFY_TIMEOUT = int(ENV.get("COSMOFY_TIMEOUT", DEFAULT_TIMEOUT))
"""Network timeout (override via environment variable)."""



#!/usr/bin/env python

# std
from __future__ import annotations
from dataclasses import dataclass
import logging
import shlex
import sys


# pkg
from cosmofy.args import common_args
from cosmofy.args import CommonArgs
from cosmofy.args import get_banner
from cosmofy.args import global_options
from cosmofy.baton import arg
from cosmofy.baton import Command
from cosmofy.zipfile2 import ZipFile2

from .rm import remove_path


log = logging.getLogger(__name__)

usage = f"""\
Get or set the special `.args` files in a Cosmopolitan bundle.

These are the arguments to the Cosmopolitan Python.

Usage: cosmofy fs args <BUNDLE> [OPTIONS] [VAL]

Arguments:
{common_args}
  [VAL]                     value to set (if omitted, current value is printed)

{global_options}
"""


@dataclass
class Args(CommonArgs):
    __doc__ = usage
    val: str = arg("", positional=True, required=False)


def get_args(bundle: ZipFile2) -> str:
    """Return a single-line representation of the `.args` file."""
    name = ".args"
    info = bundle.NameToInfo.get(name)
    if info is None:
        return ""
    return shlex.join(bundle.read(".args").decode("utf-8").split("\n"))


def set_args(bundle: ZipFile2, val: str, *, dry_run: bool = False) -> str:
    """Set the value of the `.args` file."""
    banner = get_banner(dry_run)
    for_real = not dry_run

    val = val.strip()
    if bundle.NameToInfo.get(".args") is not None:
        remove_path(bundle, ".args", force=True, dry_run=dry_run)
    if for_real:
        bundle.add_file(".args", "\n".join(shlex.split(val)))
    log.info(f"{banner}set `.args` to '{val}'")
    return val


def run(args: Args) -> int:
    """Entry point for `cosmofy fs args`."""
    args.setup_logger()
    try:
        assert args.ensure_bundle() and args.bundle
        with ZipFile2(args.bundle, "a") as bundle:
            if args.val:
                set_args(bundle, args.val, dry_run=args.dry_run)
            else:
                print(get_args(bundle), flush=True)  # don't log!
    except Exception as e:
        args.show_error(log, e)
        return 2
    return 0


cmd = Command("cosmofy.fs.args", Args, run)

if __name__ == "__main__":
    sys.exit(cmd.main())



#!/usr/bin/env python

# std
from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
import logging
import sys
import os

# pkg
from cosmofy.args import common_args
from cosmofy.args import CommonArgs
from cosmofy.args import get_banner
from cosmofy.args import global_options
from cosmofy.baton import arg
from cosmofy.baton import Command
from cosmofy.zipfile2 import ZipFile2

from .rm import remove_path


log = logging.getLogger(__name__)

usage = f"""\
Add files to a Cosmopolitan bundle.

Usage: cosmofy fs add <BUNDLE> [OPTIONS] <FILE>...

Arguments:
{common_args}
  <FILE>...                 files relative to current directory to add

  tip: Use `--` to separate options from filenames that start with `-`
  Example: cosmofy fs add bundle.zip -- -weird-filename.txt

Options:
  -f, --force               overwrite existing files
      --chdir <PATH>        change to this directory before adding
      --dest                prefix to add in the bundle
                            Most python packages go into `Lib/site-packages`

{global_options}
"""


@dataclass
class Args(CommonArgs):
    __doc__ = usage
    file: list[str] = arg(list, positional=True, required=True, action="extend")
    chdir: Path | None = arg(None)
    dest: str = arg("")
    force: bool = arg(False, short="-f")
    # compile_bytecode: bool = arg(False)


def sanitize_zip_path(dest: str) -> str:
    """Normalize and validate a zip archive path.

    - Converts backslashes to forward slashes
    - Removes leading slashes (no absolute paths)
    - Rejects path traversal attempts (`..`)

    Raises:
        ValueError: If path contains `..` segments
    """
    if "\x00" in dest:
        raise ValueError(f"refusing path with null byte: {dest!r}")

    clean = dest.replace("\\", "/")
    if clean.startswith("/"):
        log.warning(f"Stripping leading '/' from absolute path: {dest}")

    parts = [p for p in clean.split("/") if p and p != "."]
    if any(p == ".." for p in parts):
        raise ValueError(f"refusing path with '..': {dest}")
    return "/".join(parts)


def add_data(
    bundle: ZipFile2,
    data: str | bytes | bytearray,
    dest: str,
    *,
    force: bool = False,
    # global
    dry_run: bool = False,
    level: int = logging.INFO,
) -> None:
    """Add `data` to `bundle` at location `dest`."""
    dest = sanitize_zip_path(dest)
    banner = get_banner(dry_run)
    for_real = not dry_run

    if bundle.NameToInfo.get(dest) is not None:  # already exists
        if force:
            remove_path(
                bundle,
                dest,
                force=force,
                recursive=False,
                dry_run=dry_run,
                level=level,
            )
        else:
            err = f"{banner}file already exists: {dest}"
            err += "\n  tip: use --force to overwrite it"
            raise FileExistsError(err)

    if for_real:
        bundle.add_file(dest, data)
    log.log(level, f"{banner}added {dest}")


def add_path(
    bundle: ZipFile2,
    src: Path,
    dest: str,
    *,
    force: bool = False,
    # global
    dry_run: bool = False,
    level: int = logging.INFO,
    _seen: set[str] | None = None,
) -> None:
    """Add `src` to `bundle` at location `dest`."""
    _seen = _seen or set()
    real = str(src.resolve())
    if real in _seen:  # already done
        log.warning(f"skipping circular symlink: {src}")
        return
    _seen.add(real)

    dest = sanitize_zip_path(dest)
    banner = get_banner(dry_run)
    if not src.exists():
        raise FileNotFoundError(f"{banner}cannot find file: {src.resolve()}")

    if src.is_file():
        add_data(
            bundle, src.read_bytes(), dest, force=force, dry_run=dry_run, level=level
        )
    elif src.is_dir():
        for item in src.iterdir():
            add_path(
                bundle,
                item,
                dest + f"/{item.name}",
                force=force,
                dry_run=dry_run,
                level=level,
                _seen=_seen,
            )


def add_files(bundle: ZipFile2, args: Args) -> None:
    """Add files to the bundle."""
    original = Path.cwd()
    if args.chdir:
        log.debug(f"change directory: {args.chdir}")
        os.chdir(args.chdir)

    try:
        root = Path.cwd().resolve()
        for name in args.file:
            path = Path(name)
            try:
                rel = path.resolve().relative_to(root)
                dest = str(Path(args.dest) / rel) if args.dest else str(rel)
            except ValueError:
                # Path is not under root, use as-is
                dest = args.dest + name if args.dest else name
            add_path(bundle, Path(name), dest, force=args.force, dry_run=args.dry_run)
    finally:
        if args.chdir:
            log.debug(f"change directory: {original}")
            os.chdir(original)


def run(args: Args) -> int:
    """Entry point for `cosmofy fs add`."""
    args.setup_logger()
    try:
        assert args.ensure_bundle() and args.bundle

        args.dest = args.dest.strip("/")
        # NOTE: We remove the leading slash so that "/Lib" properly
        # becomes "Lib". We remove the trailing slash so we can
        # add bits properly.

        # good to go
        with ZipFile2(args.bundle, mode="a") as bundle:
            add_files(bundle, args)
    except Exception as e:
        args.show_error(log, e)
        return 2
    return 0


cmd = Command("cosmofy.fs.add", Args, run, usage)

if __name__ == "__main__":
    sys.exit(cmd.main())



"""Enhanced `ZipFile`"""

# std
from __future__ import annotations
from datetime import datetime
from fnmatch import fnmatch
from operator import attrgetter
from zipfile import ZIP_DEFLATED
from zipfile import ZipFile
from zipfile import ZipInfo
import logging

log = logging.getLogger(__name__)


class ZipFile2(ZipFile):
    """Extension of `zipfile.ZipFile` that allows removing members.

    Note: This class is not safe for concurrent access from multiple
    processes. Callers should ensure exclusive access to the archive.
    """

    _writing: bool

    # NOTE: This function only works on python >= 3.11
    # def add_dir(self, name: str, mode: int = 0o777, date: datetime = now) -> ZipFile2:
    #     """Add a directory to an archive with permissions."""
    #     if not name.endswith("/"):
    #         name += "/"
    #     info = ZipInfo(name, date.timetuple()[:6])
    #     info.compress_size = 0
    #     info.CRC = 0
    #     info.external_attr = ((0o40000 | mode) & 0xFFFF) << 16
    #     info.external_attr |= 0x10
    #     info.file_size = 0
    #     info.flag_bits |= 0x800
    #     self.mkdir(info)
    #     return self

    def add_file(
        self,
        path: str,
        data: bytearray | bytes | str,
        mode: int = 0o644,
        date: datetime | None = None,
    ) -> ZipFile2:
        """Add a file to an archive with appropriate permissions."""
        if date is None:
            date = datetime.now()

        info = ZipInfo(path, date.timetuple()[:6])
        info.compress_type = ZIP_DEFLATED
        info.external_attr = (0x8000 | (mode & 0xFFFF)) << 16
        info.flag_bits |= 0x800
        self.writestr(info, data)
        return self

    # https://github.com/python/cpython/commit/659eb048cc9cac73c46349eb29845bc5cd630f09
    def remove(self, member: str | ZipInfo) -> ZipFile2:
        """Remove a file from the archive. The archive must be open with mode 'a'"""
        if self.mode != "a":
            raise RuntimeError("remove() requires mode 'a'")
        if not self.fp:
            raise ValueError("Attempt to write to ZIP archive that was already closed")
        if self._writing:
            raise ValueError(
                "Can't write to ZIP archive while an open writing handle exists."
            )

        # zinfo
        if isinstance(member, ZipInfo):
            return self._remove_member(member)

        # name
        if "*" not in member and "?" not in member:
            return self._remove_member(self.getinfo(member))

        # glob
        for item in self.filelist:
            if fnmatch(item.filename, member):
                self._remove_member(item)
        return self

    def _remove_member(self, member: ZipInfo) -> ZipFile2:
        """Internal method to remove a member."""
        fp = self.fp
        assert fp

        # sort by header_offset in case central dir has different order
        entry_offset = 0
        filelist = sorted(self.filelist, key=attrgetter("header_offset"))
        last_index = len(filelist) - 1
        for i, info in enumerate(filelist):
            if info.header_offset < member.header_offset:  # keep going until target
                continue

            # get the total size of the entry
            entry_size = None
            if i == last_index:
                entry_size = self.start_dir - info.header_offset
            else:
                entry_size = filelist[i + 1].header_offset - info.header_offset

            if member == info:  # set the entry offset
                entry_offset = entry_size
                continue
            # move all subsequent entries

            # read the actual entry data
            fp.seek(info.header_offset)
            entry_data = fp.read(entry_size)

            # update the header
            info.header_offset -= entry_offset

            # write the entry to the new position
            fp.seek(info.header_offset)
            fp.write(entry_data)
            fp.flush()

        # update state
        self.start_dir -= entry_offset
        self.filelist.remove(member)
        del self.NameToInfo[member.filename]
        self._didModify = True

        # seek to the start of the central dir
        fp.seek(self.start_dir)
        return self

    def _write_end_record(self) -> None:
        """Write the end record to the file and truncate extra space."""
        super()._write_end_record()  # type: ignore
        if self.fp and hasattr(self.fp, "truncate"):
            self.fp.truncate()
        else:  # pragma: no cover
            # This is hard to test without messing up other things.
            print(
                "WARNING: truncate unimplemented, zip WILL be corrupted if you removed a member!"
            )



#!/usr/bin/env python
"""cosmofy: Cosmopolitan Python bundler"""

from __future__ import annotations
from dataclasses import dataclass
import logging
import sys

# pkg
from .args import global_options
from .args import GlobalArgs
from .args import log_normal
from .baton import arg
from .baton import ColorFormatter
from .baton import ColorHandler
from .baton import Command
from .bundle import cmd as bundle
from .fs.__main__ import cmd as fs
from .self.__main__ import cmd as self
from .self.version import print_version
from .updater.__main__ import cmd as updater


log = logging.getLogger(__name__)

usage = f"""\
A Cosmopolitan Python bundler.

Usage: cosmofy [OPTIONS] <COMMAND>

Commands:
  bundle                    build and bundle a project
  updater                   install/uninstall bundle self-updater
  fs                        inspect and modify an existing bundle
  self                      manage the `cosmofy` executable

Options:
      --version             display the program version and exit

{global_options}
"""


def setup_logger() -> None:
    """Setup root logger, if needed."""
    root = logging.getLogger()
    if not root.handlers:  # pragma: no cover
        handler = ColorHandler()
        handler.setFormatter(ColorFormatter(log_normal))
        logging.basicConfig(level=logging.INFO, handlers=[handler])


@dataclass
class Args(GlobalArgs):
    command: str = arg("", positional=True)  # optional so we can show usage
    version: bool = arg(False)


def run(args: Args) -> int:
    """Main entry point for `cosmofy`."""
    args.setup_logger()
    if args.version:
        print_version()
        return 0

    # NOTE: only called when there was no subcommand found
    cmd.show_usage()
    return 0


cmd = Command(
    "cosmofy",
    Args,
    run,
    usage,
    {
        "bundle": bundle,
        "updater": updater,
        "fs": fs,
        "self": self,
    },
)


def main() -> None:
    """Main entry point for console scripts."""
    setup_logger()
    sys.exit(cmd.main())


if __name__ == "__main__":  # pragma: no cover
    main()



#!/usr/bin/env python

# std
from __future__ import annotations
from dataclasses import dataclass
from importlib.metadata import version
from typing import Literal
import json
import logging
import sys

# pkg
from cosmofy.args import global_options
from cosmofy.args import GlobalArgs
from cosmofy.baton import arg
from cosmofy.baton import COLOR_MODE
from cosmofy.baton import Command
from cosmofy.baton import render_tags


log = logging.getLogger(__name__)

usage = f"""\
Display `cosmofy`'s version.

Usage: cosmofy self version [OPTIONS]

Options:
      --short               only print the version
      --output-format NAME  [default: text][choices: text, json]

{global_options}
"""

OUTPUT_FORMAT = Literal["text", "json"]


@dataclass
class Args(GlobalArgs):
    __doc__ = usage
    short: bool = arg(False)
    output_format: OUTPUT_FORMAT = arg("text")


def get_version(pkg: str = "cosmofy") -> dict[str, str]:
    """Return package metadata."""
    return {"package_name": pkg, "version": version(pkg)}


def print_version(
    *,
    short: bool = False,
    output_format: OUTPUT_FORMAT = "text",
    color: COLOR_MODE = "auto",
) -> str:
    """Print version information and return the text printed."""
    result = ""
    data = get_version()
    if output_format == "json":
        result = json.dumps(data, indent=2)
    else:
        result = f"[cyan]{data['version']}[/]"
        if not short:
            result = f"{data['package_name']} {result}"
        result = render_tags(result, color=color)

    print(result)  # do not log
    return result


def run(args: Args) -> int:
    """Entry point for `cosmofy self version`."""
    args.setup_logger()
    try:
        print_version(
            short=args.short,
            output_format=args.output_format,
            color=args.color,
        )
    except Exception as e:
        args.show_error(log, e)
        return 2
    return 0


cmd = Command("cosmofy.self.version", Args, run)

if __name__ == "__main__":
    sys.exit(cmd.main())



#!/usr/bin/env python
"""pythonoid: small simulated python CLI"""

# std
from __future__ import annotations
from importlib.machinery import SourceFileLoader
from importlib.util import MAGIC_NUMBER
from os import environ as ENV
from pathlib import Path
import code as repl
import dataclasses
import logging
import marshal
import re
import runpy
import sys
import traceback

log = logging.getLogger(__name__)


Pkg = tuple[str, ...]
"""Package information."""

MODULE_SUFFIXES = (".py", ".pyc")
"""Python module suffixes."""

PACKAGE_STEMS = ("__init__", "__main__")
"""File stems that indicate a python package."""

PACKAGE_FILES = tuple(p + s for p in PACKAGE_STEMS for s in MODULE_SUFFIXES)
"""File names that indicate a python package."""

MAIN_FILES = ("__main__.py", "__main__.pyc")
"""File names that indicate python package has a main."""

RE_MAIN = re.compile(
    rb"""
    (^|\n)if\s*(
    __name__\s*==\s*['"]__main__['"]| # written the normal way
    ['"]__main__['"]\s*==\s*__name__) # written in reverse
    """,
    re.VERBOSE,
)
"""Regex for detecting a main section in `bytes`."""

PYTHON_CALL = "-c 'import sys; import {pkg} as _1; sys.exit(_1.{fn})'"
"""Format for a python call."""


# https://github.com/metaist/ds/blob/main/src/ds/parsers/pyproject_rye.py#L153C1-L174C46
def python_call(call: str) -> str:
    """Return python arguments a formatted `call` string.

    See: https://rye.astral.sh/guide/pyproject/#call

    >>> python_call("http.server")
    '-m http.server'

    >>> python_call("builtins:help") == PYTHON_CALL.format(pkg="builtins", fn="help()")
    True

    >>> python_call("builtins:print('Hello World!')") == PYTHON_CALL.format(
    ...     pkg="builtins", fn="print('Hello World!')")
    True
    """
    if ":" not in call:
        return f"-m {call}"

    pkg, fn = call.split(":", 1)
    if not fn.endswith(")"):
        fn = f"{fn}()"
    return PYTHON_CALL.format(pkg=pkg, fn=fn)


# https://github.com/python/cpython/blob/3.12/Lib/importlib/_bootstrap_external.py#L79C1-L81C55
def _pack_uint32(x: int | float) -> bytes:
    """Convert a 32-bit integer to little-endian."""
    return (int(x) & 0xFFFFFFFF).to_bytes(4, "little")


def compile_python(path: Path, source: bytes | None = None) -> bytearray:
    """Return the bytecode."""
    source = path.read_bytes() if source is None else source
    stats = path.stat()
    mtime = stats.st_mtime
    source_size = stats.st_size

    # https://github.com/python/cpython/blob/3.12/Lib/importlib/_bootstrap_external.py#L1059
    code = compile(source, path, "exec", dont_inherit=True, optimize=-1)

    # https://github.com/python/cpython/blob/3.12/Lib/importlib/_bootstrap_external.py#L764
    data = bytearray(MAGIC_NUMBER)
    data.extend(_pack_uint32(0))
    data.extend(_pack_uint32(mtime))
    data.extend(_pack_uint32(source_size))
    data.extend(marshal.dumps(code))
    return data


# https://github.com/python/cpython/blob/32119fc377a4d9df524a7bac02b6922a990361dd/Python/initconfig.c#L233
USAGE = f"""\
usage: {Path(__file__).name} [option] ... [-c cmd | -m mod | file | -] [arg] ...
Options (and corresponding environment variables):
-c cmd : program passed in as string (terminates option list)
-h     : print this help message and exit (also -? or --help)
-i     : inspect interactively after running script; forces a prompt even
         if stdin does not appear to be a terminal; also PYTHONINSPECT=x
-I     : isolate Python from the user's environment (implies -E and -s)
-m mod : run library module as a script (terminates option list)
-q     : don't print version and copyright messages on interactive startup
-V     : print the Python version number and exit (also --version)
         when given twice, print more information about the build

Arguments:
file   : program read from script file
-      : program read from stdin (default; interactive mode if a tty)
arg ...: arguments passed to program in sys.argv[1:]

Python has many more options that pythonoid does not support:
https://docs.python.org/3/using/cmdline.html"""


@dataclasses.dataclass
class PythonArgs:
    """Subset of python command-line arguments.

    See: https://docs.python.org/3/using/cmdline.html
    """

    c: str | None = None
    """Command to execute."""

    h: bool = False
    """Show help."""

    i: bool = ENV.get("PYTHONINSPECT", "") == "x"
    """Interactive mode."""

    m: str | None = None
    """Module name."""

    q: bool = False
    """No copyright and version messages."""

    V: bool = False
    """Version information."""

    VV: bool = False
    """Verbose version information."""

    script: str | None = None
    """Script to execute."""

    argv: list[str] = dataclasses.field(default_factory=list)
    """Remaining arguments to <script>, <module>, or <command>."""

    @staticmethod
    def parse(argv: list[str]) -> PythonArgs:
        """Parse a subset of python command-line args."""
        args = PythonArgs()
        UNSUPPORTED = """
            --help-env --help-xoptions --help-all
            -b -B --check-hash-based-pycs -d -E -I -O -OO -P -R
            -s -S -u -v -W -x -X
        """.split()
        while argv:
            arg = argv.pop(0)
            if (
                arg.startswith("-") and not arg.startswith("--") and len(arg) > 2
            ):  # expand
                argv = [f"-{a}" for a in arg[1:]] + argv
                continue

            if arg in ["-c", "-m"]:  # flags with an argument
                if not argv:
                    raise ValueError(f"Argument expected for the {arg} option")
                setattr(args, arg[1:], argv.pop(0))
                args.argv += argv  # whatever is left
                # NOTE: for -m it should be the full path to the module
                args.argv.insert(0, arg)
                break  # remainder are argv
            elif arg in ["-?", "-h", "--help"]:
                args.h = True
            elif arg in ["-i", "-q"]:  # <bool> flags
                setattr(args, arg[1:], True)
            elif arg in ["-"]:
                args.c = sys.stdin.read()
                if sys.stdin.isatty():
                    args.i = True
                args.argv += argv  # whatever is left
                args.argv.insert(0, "-")
                break  # remainder are argv
            elif arg in ["-V", "--version"]:
                if args.V:
                    args.VV = True
                else:
                    args.V = True
            elif arg in UNSUPPORTED:
                raise ValueError(f"Unsupported (but valid) python option: {arg}")
            elif arg.startswith("-"):
                raise ValueError(f"Unknown option: {arg}")
            else:  # <script>
                args.script = arg
                args.argv += argv
                args.argv.insert(0, arg)  # "script name as given on the command line"
                break  # remainder are argv

        if not any([args.c, args.h, args.i, args.m, args.V, args.script]):
            args.i = True
            args.argv.insert(0, "")
        return args


def run_python(argv: list[str]) -> int:
    """Simulate running python with the given args."""
    try:
        args = PythonArgs.parse(argv)
    except ValueError as e:
        log.error(e)
        return 2

    if args.h:
        print(USAGE)
        return 0

    if args.V:
        version = sys.version
        if not args.VV:
            version = ".".join(str(x) for x in sys.version_info[:3])
        print(f"Python {version}")
        return 0

    if isinstance(__builtins__, dict):
        loader = __builtins__["__loader__"]
    else:  # pragma: no cover
        # During testing, __builtins__ is a dict.
        loader = __builtins__.__loader__

    local: dict[str, object] = {
        "__name__": "__main__",
        "__doc__": None,
        "__package__": None,
        "__loader__": loader,
        "__spec__": None,
        "__annotations__": {},
        "__builtins__": __builtins__,
    }

    sys.argv = args.argv  # NOTE: intentionally destructive
    code = 0
    try:
        if args.c:  # execute in the context of the locals
            args.q = True
            exec(args.c, local, local)
        elif args.m:
            args.q = True
            runpy.run_module(args.m, local, "__main__", alter_sys=True)
        elif args.script:
            args.q = True
            local["__loader__"] = SourceFileLoader
            runpy.run_path(args.script, local, "__main__")
    except Exception as e:
        code = 1
        # NOTE: We skip the calling frame to emulate the CLI better.
        tb = sys.exc_info()[2]
        tb_next = tb.tb_next if tb else tb
        print("".join(traceback.format_exception(e.__class__, e, tb_next)), end="")

    if args.i:  # after <command>, <module>, or <script>, enter interactive mode
        code = 0
        info = 'Type "help", "copyright", "credits" or "license" for more information.'
        banner = f"Python {sys.version} on {sys.platform}\n{info}"
        if args.q:
            banner = ""
        repl.interact(banner=banner, local=local, exitmsg="")

    return code


if __name__ == "__main__":  # pragma: no cover
    sys.exit(run_python(sys.argv[1:]))



#!/usr/bin/env python

# std
from __future__ import annotations
from dataclasses import dataclass
from urllib.error import HTTPError
from zipfile import ZipFile
import json
import logging
import sys

# pkg
from cosmofy.args import common_args
from cosmofy.args import CommonArgs
from cosmofy.args import get_banner
from cosmofy.args import global_options
from cosmofy.baton import arg
from cosmofy.baton import Command

from .add import PATH_RECEIPT
from .add import RECEIPT_URL
from .receipt import Receipt


log = logging.getLogger(__name__)


usage = f"""\
EXPERIMENTAL: Check if a cosmofy bundle has an update.

Usage: cosmofy updater check <BUNDLE> [OPTIONS]

Arguments:
{common_args}

Options:
      --receipt-url <URL>   override the published receipt URL
                            [env: RECEIPT_URL={RECEIPT_URL}]

{global_options}
"""


@dataclass
class Args(CommonArgs):
    __doc__ = usage
    receipt_url: str = arg(RECEIPT_URL)


def get_local_receipt(bundle: ZipFile) -> Receipt:
    """Return `Receipt` embedded in `bundle`."""
    if PATH_RECEIPT not in bundle.NameToInfo:
        raise FileNotFoundError("cannot find cosmofy receipt in bundle")
    return Receipt.from_dict(json.loads(bundle.read(PATH_RECEIPT)))


def get_remote_receipt(url: str, *, dry_run: bool = False) -> Receipt:
    """Try to download a receipt."""
    banner = get_banner(dry_run)
    log.info(f"{banner}download: {url}")
    try:
        if dry_run:
            return Receipt()
        else:
            return Receipt.from_url(url)
    except HTTPError as e:
        log.error(f"{e}: {url}")
        raise FileNotFoundError("cannot download published receipt") from e


def check(
    bundle: ZipFile, *, receipt_url: str = "", dry_run: bool = False
) -> tuple[bool, Receipt, Receipt]:
    """Return published receipt after checking if there are is a newer version."""
    banner = get_banner(dry_run)
    local = get_local_receipt(bundle)
    log.debug(f"{banner}embedded receipt: {local}")

    url = receipt_url or local.receipt_url
    log.debug(f"{banner}receipt URL: {url}")

    remote = get_remote_receipt(url, dry_run=dry_run)
    log.debug(f"{banner}published receipt: {remote}")

    is_newer = remote.is_newer(local)
    if is_newer:
        log.info(f"{banner}new version found: {remote.version} ({remote.date})")
    else:
        log.info(
            f"{banner}[bold green]success[/]: you have the latest version ([bold cyan]{remote.version}[/])"
        )
    return is_newer, local, remote


def run(args: Args) -> int:
    """Main entry point for `cosmofy updater check`."""
    args.setup_logger()
    log.warning("EXPERIMENTAL: `cosmofy updater` operations are experimental")
    try:
        assert args.ensure_bundle() and args.bundle
        check(
            ZipFile(args.bundle, "r"),
            receipt_url=args.receipt_url,
            dry_run=args.dry_run,
        )
    except Exception as e:
        args.show_error(log, e)
        return 2
    return 0


cmd = Command("cosmofy.updater.check", Args, run)

if __name__ == "__main__":
    sys.exit(cmd.main())



# std
from fnmatch import translate
from typing import Iterator
import re


def shell_match(name: str, pat: str) -> bool:
    """Match following shell glob rules.
    - `*` doesn't match `/` or leading dot
    - `**` matches anything
    """
    if name.startswith(".") and not pat.startswith("."):
        return False

    # Protect ** patterns with placeholders
    pat = pat.replace("**/", "\x00")
    pat = pat.replace("**", "\x01")
    regex = translate(pat)
    # * should not match /
    regex = regex.replace(".*", "[^/]*")
    # * after / should not match leading dot (hidden files)
    regex = regex.replace("/[^/]*", "/(?![.])[^/]*")
    # **/ matches zero or more non-hidden directories
    regex = regex.replace("\x00", "((?![.])[^/]*/)*")
    # ** at end matches anything
    regex = regex.replace("\x01", ".*")
    return re.match(regex, name) is not None


def expand_glob(names: list[str], pat: str) -> Iterator[str]:
    """Resolve a file pattern.

    Glob patterns are matched against `names`. Literal paths are passed
    through unchanged (caller is responsible for checking existence).
    """
    if "*" in pat or "?" in pat:
        yield from (name for name in names if shell_match(name, pat))
    else:
        yield pat
